/*
** glhw.c
**
** Written by Ian Buck
** 6/15/2000
** http://graphics.stanford.edu/software/wiregl
**
** Copyright  2000 The Board of Trustees of The Leland Stanford Junior
** University. All rights reserved.
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or modify
** this software, provided that the above copyright notice and this
** permission notice appear in all copies of this software.  No part of
** this software or any derivatives thereof may be used in graphics systems
** for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind, 
** express, implied or otherwise, including without limitation, any warranty 
** of merchantability or fitness for a particular purpose.
**
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "glcontext.h"
#include "glerror.h"

#ifdef WINDOWS
#include <windows.h>
#elif defined(IRIX) || defined(IRIX64) || defined(Linux)
#include <dlfcn.h>
#else
#error DSO
#endif

#define GLHWAPI_DEFINE
#include "glhw.h"

#if defined( WINDOWS )
#define SYSTEM_GL_NAME "opengl32"
#else
#define SYSTEM_GL_NAME "libGL.so"
#endif

#define MAXAPI 400
void **APIbase[MAXAPI+1] = {NULL};
void **APIbeginendbase[MAXAPI+1] = {NULL};
void **APIvertexbase[MAXAPI+1] = {NULL};
void **APIvertexbase_st[MAXAPI+1] = {NULL};
void **APIarrayelementbase;

void *APIpk[MAXAPI+1] = {NULL};
void *APIst[MAXAPI+1] = {NULL};
void *APIbeginendpk[MAXAPI+1] = {NULL};
void *APIbeginendst[MAXAPI+1] = {NULL};
void *APIvertexpack[MAXAPI+1] = {NULL};
void *APIvertexBBpack[MAXAPI+1] = {NULL};
void *APIvertexCpack[MAXAPI+1] = {NULL};

void *APIarrayelement_locked = (void *) __glstate_ArrayElement_locked;
void *APIarrayelement_unlocked = (void *) __glstate_ArrayElement;

int APIcount = 0;

/* These are defined below */
extern GLjumptable hw_api[];

extern GLapi nop_api[];
extern GLapi glcontext_api[];
extern GLapi system_api[];

extern const char packonly[][64];
extern const char stateonly[][64];
extern const char beginend[][64];
extern const char vertexfuncs[][64];

/* LoadApi Functions
** Used during stream processing
** Note: These reimplemented to be a lot faster
** since the funcptrs are stored in a linear
** array, we could just do a memcopy. However,
** the function groups cannot cause non-linear
** groupings.  (I think this isn't a problem
** currently).
*/ 

void
__glhw_LoadPackApi(void) {
	void ***base = APIbase;
	void **pk = APIpk;
	while (*base)
		**base++ = *pk++;
}

void
__glhw_LoadStateApi(void) {
	void ***base = APIbase;
	void **st = APIst;
	while (*base)
		**base++ = *st++;
}

void
__glhw_LoadBeginEndApi(void) {
	void ***base = APIbeginendbase;
	void **pk = APIbeginendpk;
	while (*base)
		**base++ = *pk++;
}

void
__glhw_UnloadBeginEndApi(void) {
	void ***base = APIbeginendbase;
	void **st = APIbeginendst;
	while (*base)
		**base++ = *st++;
}

void
__glhw_LoadVertexPackApi(void) {
	void ***base = APIvertexbase;
	void ***basest = APIvertexbase_st;
	void **pk = APIvertexpack;
    while (*base) {
		**base++ = *pk;
		**basest++ = *pk++;
	}
}

void
__glhw_LoadVertexBBPackApi(void) {
	void ***base = APIvertexbase;
	void ***basest = APIvertexbase_st;
	void **pk = APIvertexBBpack;
    while (*base) {
		**base++ = *pk;
		**basest++ = *pk++;
	}
}

void
__glhw_LoadVertexCPackApi(void) {
	void ***base = APIvertexbase;
	void ***basest = APIvertexbase_st;
	void **pk = APIvertexCpack;
    while (*base) {
		**base++ = *pk;
		**basest++ = *pk++;
	}
}

void
__glhw_LockArrays() {
	*APIarrayelementbase = APIarrayelement_locked;
}

void
__glhw_UnlockArrays() {
	*APIarrayelementbase = APIarrayelement_unlocked;
}


/* glhw init function*/
void
__glhw_RegisterApi(GLconfig *c) {
	int i, j, beginend_i=0, vertex_i=0;
	GLapi *p;

	GLjumptable *base = c->immediateapi;
	GLapi *pk = c->packapi;
	GLapi *st = glcontext_api;

	if (!base || !base->func) return;

	/* if they don't give us a pack api,
	** use the systems.
	*/
	if (!pk || !pk->func) {
		__glhw_LoadSystemApi(system_api);
		pk = system_api;
	}

	for (i=0; i < MAXAPI && APIbase[i]; i++);

	for (; i < MAXAPI && *base->name; i++, base++) {
		APIbase[i] = base->func;
		APIpk[i] = (void *) __glhw_error;
		APIst[i] = (void *) __glhw_error;

		/* Assign the hwerror */
		for (p=hwerror; *p->name; p++) {
			if (!strcmp(base->name, p->name)) {
				APIpk[i] = p->func;
				APIst[i] = p->func;
			}
		}

		/* Assign Pack Function */
		for (p=pk; *p->name; p++) {
			if (!strcmp(base->name, p->name)) {
				APIpk[i] = p->func;
				break;
			}
		}

		/* Assign State Function */
		for (p=st; *p->name; p++) {
			if (!strcmp(base->name, p->name)) {
				APIst[i] = p->func;
				/* Patch for allowing different ArrayElement call */
				if (!strcmp(base->name, "glArrayElement") &&
					c->resolve_arrays)
					APIst[i] = (void *) __glstate_ArrayElement_resolve;
				break;
			}
		}

		/* Check Pack only list */
		for (j=0; *packonly[j]; j++) {
			if (!strcmp(packonly[j], base->name)) {
				APIst[i] = APIpk[i];
				break;
			}
		}

		/* Check State only list */
		for (j=0; *stateonly[j]; j++) {
			if (!strcmp(stateonly[j], base->name)) {
				APIpk[i] = APIst[i];
				break;
			}
		}

		/* ArrayElement (Unimplemented)
**		if (!strcmp(base->name, "ArrayElement")) {
**			APIarrayelementbase = base->func;
**		}
**/

		/* Update Vertex Pack APIs */
		for (j=0; *vertexfuncs[j]; j++) {
			if (!strcmp(vertexfuncs[j], base->name)) {
				APIvertexbase[vertex_i] = base->func;
				APIvertexbase_st[vertex_i] = APIst+i;
				APIvertexpack[vertex_i] = APIpk[i];
				if (!c->vertexBBapi) {
					APIvertexBBpack[vertex_i] = APIpk[i];
				} else
					for (j=0; *(c->vertexBBapi[j].name); j++)
						if (!strcmp(c->vertexBBapi[j].name, base->name)) {
							APIvertexBBpack[vertex_i] = c->vertexBBapi[j].func;
							APIst[i]= c->vertexBBapi[j].func;  //By Default????
							break;
						}
				if (!c->vertexCapi) {
					APIvertexCpack[vertex_i] = APIpk[i];
				} else
					for (j=0; *(c->vertexCapi[j].name); j++)
						if (!strcmp(c->vertexCapi[j].name, base->name)) {
							APIvertexCpack[vertex_i] = c->vertexCapi[j].func;
							break;
						}
				vertex_i++;
				break;
			}
		}

		/* Update Begin End APIs */
		for (j=0; *beginend[j]; j++) {
			if (!strcmp(beginend[j], base->name)) {
				APIbeginendbase[beginend_i] = base->func;
				APIbeginendpk[beginend_i] = APIpk[i];
				APIbeginendst[beginend_i] = APIst[i];
				beginend_i++;
				break;
			}
		}
	}

	/* All lists end in a null */
	APIbase[i] = NULL;
	APIpk[i] = NULL;
	APIst[i] = NULL;
	APIbeginendbase[beginend_i] = NULL;
	APIbeginendpk[beginend_i] = NULL;
	APIbeginendst[beginend_i] = NULL;
	APIvertexbase[vertex_i] = NULL;
	APIvertexbase_st[vertex_i] = NULL;
	APIvertexpack[vertex_i] = NULL;
	APIvertexBBpack[vertex_i] = NULL;
	APIvertexCpack[vertex_i] = NULL;
	APIcount = i;
}

void __glhw_LoadSystemApi (GLapi *api) {
	int i,j;
#if defined(WINDOWS)
	char name[256];
#endif

	for (i=0; api[i].func; i++)
		api[i].func = NULL;

#if defined(WINDOWS)
	GetSystemDirectory(name, 256);
	strcat (name, "\\opengl32.dll");
	HMODULE hinstLib = LoadLibrary( name );
	for (i=0; *api[i].name; i++)
		api[i].func = (void *) GetProcAddress(hinstLib, api[i].name);

#elif defined(IRIX) || defined(IRIX64) || defined(Linux)
	void *hinstLib = dlopen("libGL.so", RTLD_LAZY );
	for (i=0; *api[i].name; i++)
		api[i].func = (void *) dlsym(hinstLib, api[i].name);

#else
#error DSO Loading not implemented for this OS.
#endif
	
	for (i=0; *api[i].name; i++)
		if (api[i].func == NULL)
			for (j=0; nop_api[j].func; j++)
				if (!strcmp (api[i].name, nop_api[j].name)) {
					api[i].func = nop_api[j].func;
					break;
				}
}

void __glhw_LoadSwitchApi (GLapi *api) {
	int i,j;

	if (!api) {
		__glhw_LoadSystemApi(system_api);
		api = system_api;
	}

	/* Initialize everybody to NULL */
	for (i=0; *hw_api[i].name; i++)
		*((void **) hw_api[i].func) = NULL;

	/* Load all api */
	for (i=0; *hw_api[i].name; i++)
		for (j=0; *api[j].name; j++)
			if (!strcmp (hw_api[i].name, api[j].name)) {
				*((void **) hw_api[i].func) = api[j].func;
				break;
			}

	/* Load the remaining nops */
	for (i=0; *hw_api[i].name; i++)
		if (*((void **) hw_api[i].func) == NULL)
			for (j=0; nop_api[j].func; j++)
				if (!strcmp (hw_api[i].name, nop_api[j].name)) {
					*((void **) hw_api[i].func) = nop_api[j].func;
					break;
				}
}

void __glhw_error() {
	__glerror(__LINE__, __FILE__, GL_NO_ERROR,
		"Function API no implemented!");
	exit(1);
}

/* LoadVertex*Funcs functions */
const char vertexfuncs[][64] = {
	"glVertex2d",
	"glVertex2dv",
	"glVertex2f",
	"glVertex2fv",
	"glVertex2i",
	"glVertex2iv",
	"glVertex2s",
	"glVertex2sv",
	"glVertex3d",
	"glVertex3dv",
	"glVertex3f",
	"glVertex3fv",
	"glVertex3i",
	"glVertex3iv",
	"glVertex3s",
	"glVertex3sv",
	"glVertex4d",
	"glVertex4dv",
	"glVertex4f",
	"glVertex4fv",
	"glVertex4i",
	"glVertex4iv",
	"glVertex4s",
	"glVertex4sv",
	""
};

/* Functions which change from the 
** state to the pack mode in a 
** Begin End block
*/
const char beginend[][64] = {
	"glColor3b",
	"glColor3bv",
	"glColor3d",
	"glColor3dv",
	"glColor3f",
	"glColor3fv",
	"glColor3i",
	"glColor3iv",
	"glColor3s",
	"glColor3sv",
	"glColor3ub",
	"glColor3ubv",
	"glColor3ui",
	"glColor3uiv",
	"glColor3us",
	"glColor3usv",
	"glColor4b",
	"glColor4bv",
	"glColor4d",
	"glColor4dv",
	"glColor4f",
	"glColor4fv",
	"glColor4i",
	"glColor4iv",
	"glColor4s",
	"glColor4sv",
	"glColor4ub",
	"glColor4ubv",
	"glColor4ui",
	"glColor4uiv",
	"glColor4us",
	"glColor4usv",
	""
};

/* Functions which are always packed
*/
const char packonly[][64] = {
	"glBarrierCreate",
	"glBarrierDestroy",
	"glBarrierExec",
	"glCopyPixels",
	"glColor3b",
	"glColor3bv",
	"glColor3d",
	"glColor3dv",
	"glColor3f",
	"glColor3fv",
	"glColor3i",
	"glColor3iv",
	"glColor3s",
	"glColor3sv",
	"glColor3ub",
	"glColor3ubv",
	"glColor3ui",
	"glColor3uiv",
	"glColor3us",
	"glColor3usv",
	"glColor4b",
	"glColor4bv",
	"glColor4d",
	"glColor4dv",
	"glColor4f",
	"glColor4fv",
	"glColor4i",
	"glColor4iv",
	"glColor4s",
	"glColor4sv",
	"glColor4ub",
	"glColor4ubv",
	"glColor4ui",
	"glColor4uiv",
	"glColor4us",
	"glColor4usv",
	"glEdgeFlag",
	"glEdgeFlagv",
	"glEvalCoord1d",
	"glEvalCoord1dv",
	"glEvalCoord1f",
	"glEvalCoord1fv",
	"glEvalCoord2d",
	"glEvalCoord2dv",
	"glEvalCoord2f",
	"glEvalCoord2fv",
	"glEvalPoint1",
	"glEvalPoint2",
	"glFinish",
	"glFlush",
	"glIndexd",
	"glIndexdv",
	"glIndexf",
	"glIndexfv",
	"glIndexi",
	"glIndexiv",
	"glIndexs",
	"glIndexsv",
	"glIndexub",
	"glIndexubv",
	"glMaterialf",
	"glMaterialfv",
	"glMateriali",
	"glMaterialiv",
	"glNormal3b",
	"glNormal3bv",
	"glNormal3d",
	"glNormal3dv",
	"glNormal3f",
	"glNormal3fv",
	"glNormal3i",
	"glNormal3iv",
	"glNormal3s",
	"glNormal3sv",
	"glSemaphoreCreate",
	"glSemaphoreDestroy",
	"glSemaphoreP",
	"glSemaphoreV",
	"glTexCoord1d",
	"glTexCoord1dv",
	"glTexCoord1f",
	"glTexCoord1fv",
	"glTexCoord1i",
	"glTexCoord1iv",
	"glTexCoord1s",
	"glTexCoord1sv",
	"glTexCoord2d",
	"glTexCoord2dv",
	"glTexCoord2f",
	"glTexCoord2fv",
	"glTexCoord2i",
	"glTexCoord2iv",
	"glTexCoord2s",
	"glTexCoord2sv",
	"glTexCoord3d",
	"glTexCoord3dv",
	"glTexCoord3f",
	"glTexCoord3fv",
	"glTexCoord3i",
	"glTexCoord3iv",
	"glTexCoord3s",
	"glTexCoord3sv",
	"glTexCoord4d",
	"glTexCoord4dv",
	"glTexCoord4f",
	"glTexCoord4fv",
	"glTexCoord4i",
	"glTexCoord4iv",
	"glTexCoord4s",
	"glTexCoord4sv",
	""
};

/* Functions that don't go in a display list */
const char stateonly[][64] = {
	"glNewList",
	"glEndList",
	"glAreTexturesResident",
	"glColorPointer",
	"glDeleteLists",
	"glDeleteTextures",
	"glDisableClientState",
	"glEdgeFlagPointer",
	"glEnableClientState",
	"glFeedbackBuffer",
	"glGenLists",
	"glGenTextures",
	"glGetBooleanv",
	"glGetClipPlane",
	"glGetDoublev",
	"glGetError",
	"glGetFloatv",
	"glGetIntegerv",
	"glGetLightfv",
	"glGetLightiv",
	"glGetMapdv",
	"glGetMapfv",
	"glGetMapiv",
	"glGetMaterialfv",
	"glGetMaterialiv",
	"glGetPixelMapfv",
	"glGetPixelMapuiv",
	"glGetPixelMapusv",
	"glGetPointerv",
	"glGetPolygonStipple",
	"glGetString",
	"glGetTexEnvfv",
	"glGetTexEnviv",
	"glGetTexGendv",
	"glGetTexGenfv",
	"glGetTexGeniv",
	"glGetTexImage",
	"glGetTexLevelParameterfv",
	"glGetTexLevelParameteriv",
	"glGetTexParameterfv",
	"glGetTexParameteriv",
	"glHint",
	"glIndexPointer",
	"glInterleavedArrays",
	"glIsEnabled",
	"glIsList",
	"glIsTexture",
	"glNormalPointer",
	"glPixelStoref",
	"glPixelStorei",
	"glPopClientAttrib",
	"glPushClientAttrib",
	"glReadPixels",
	"glRenderMode",
	"glSelectBuffer",
	"glTexCoordPointer",
	"glVertexPointer",
	""
};

/* API for state switching */
GLjumptable hw_api[] = {
	{"glAccum", (void **) &__glhw_Accum},
	{"glAlphaFunc", (void **) &__glhw_AlphaFunc},
	{"glAreTexturesResident", (void **) &__glhw_AreTexturesResident},
	{"glArrayElement", (void **) &__glhw_ArrayElement},
	{"glBegin", (void **) &__glhw_Begin},
	{"glBindTexture", (void **) &__glhw_BindTexture},
	{"glBitmap", (void **) &__glhw_Bitmap},
	{"glBlendFunc", (void **) &__glhw_BlendFunc},
	{"glCallList", (void **) &__glhw_CallList},
	{"glCallLists", (void **) &__glhw_CallLists},
	{"glClear", (void **) &__glhw_Clear},
	{"glClearAccum", (void **) &__glhw_ClearAccum},
	{"glClearColor", (void **) &__glhw_ClearColor},
	{"glClearDepth", (void **) &__glhw_ClearDepth},
	{"glClearIndex", (void **) &__glhw_ClearIndex},
	{"glClearStencil", (void **) &__glhw_ClearStencil},
	{"glClipPlane", (void **) &__glhw_ClipPlane},
	{"glColor3b", (void **) &__glhw_Color3b},
	{"glColor3bv", (void **) &__glhw_Color3bv},
	{"glColor3d", (void **) &__glhw_Color3d},
	{"glColor3dv", (void **) &__glhw_Color3dv},
	{"glColor3f", (void **) &__glhw_Color3f},
	{"glColor3fv", (void **) &__glhw_Color3fv},
	{"glColor3i", (void **) &__glhw_Color3i},
	{"glColor3iv", (void **) &__glhw_Color3iv},
	{"glColor3s", (void **) &__glhw_Color3s},
	{"glColor3sv", (void **) &__glhw_Color3sv},
	{"glColor3ub", (void **) &__glhw_Color3ub},
	{"glColor3ubv", (void **) &__glhw_Color3ubv},
	{"glColor3ui", (void **) &__glhw_Color3ui},
	{"glColor3uiv", (void **) &__glhw_Color3uiv},
	{"glColor3us", (void **) &__glhw_Color3us},
	{"glColor3usv", (void **) &__glhw_Color3usv},
	{"glColor4b", (void **) &__glhw_Color4b},
	{"glColor4bv", (void **) &__glhw_Color4bv},
	{"glColor4d", (void **) &__glhw_Color4d},
	{"glColor4dv", (void **) &__glhw_Color4dv},
	{"glColor4f", (void **) &__glhw_Color4f},
	{"glColor4fv", (void **) &__glhw_Color4fv},
	{"glColor4i", (void **) &__glhw_Color4i},
	{"glColor4iv", (void **) &__glhw_Color4iv},
	{"glColor4s", (void **) &__glhw_Color4s},
	{"glColor4sv", (void **) &__glhw_Color4sv},
	{"glColor4ub", (void **) &__glhw_Color4ub},
	{"glColor4ubv", (void **) &__glhw_Color4ubv},
	{"glColor4ui", (void **) &__glhw_Color4ui},
	{"glColor4uiv", (void **) &__glhw_Color4uiv},
	{"glColor4us", (void **) &__glhw_Color4us},
	{"glColor4usv", (void **) &__glhw_Color4usv},
	{"glColorMask", (void **) &__glhw_ColorMask},
	{"glColorMaterial", (void **) &__glhw_ColorMaterial},
	{"glColorPointer", (void **) &__glhw_ColorPointer},
	{"glCopyPixels", (void **) &__glhw_CopyPixels},
	{"glCopyTexImage1D", (void **) &__glhw_CopyTexImage1D},
	{"glCopyTexImage2D", (void **) &__glhw_CopyTexImage2D},
	{"glCopyTexSubImage1D", (void **) &__glhw_CopyTexSubImage1D},
	{"glCopyTexSubImage2D", (void **) &__glhw_CopyTexSubImage2D},
	{"glCullFace", (void **) &__glhw_CullFace},
	{"glDeleteLists", (void **) &__glhw_DeleteLists},
	{"glDeleteTextures", (void **) &__glhw_DeleteTextures},
	{"glDepthFunc", (void **) &__glhw_DepthFunc},
	{"glDepthMask", (void **) &__glhw_DepthMask},
	{"glDepthRange", (void **) &__glhw_DepthRange},
	{"glDisable", (void **) &__glhw_Disable},
	{"glDisableClientState", (void **) &__glhw_DisableClientState},
	{"glDrawArrays", (void **) &__glhw_DrawArrays},
	{"glDrawBuffer", (void **) &__glhw_DrawBuffer},
	{"glDrawElements", (void **) &__glhw_DrawElements},
	{"glDrawPixels", (void **) &__glhw_DrawPixels},
	{"glEdgeFlag", (void **) &__glhw_EdgeFlag},
	{"glEdgeFlagPointer", (void **) &__glhw_EdgeFlagPointer},
	{"glEdgeFlagv", (void **) &__glhw_EdgeFlagv},
	{"glEnable", (void **) &__glhw_Enable},
	{"glEnableClientState", (void **) &__glhw_EnableClientState},
	{"glEnd", (void **) &__glhw_End},
	{"glEndList", (void **) &__glhw_EndList},
	{"glEvalCoord1d", (void **) &__glhw_EvalCoord1d},
	{"glEvalCoord1dv", (void **) &__glhw_EvalCoord1dv},
	{"glEvalCoord1f", (void **) &__glhw_EvalCoord1f},
	{"glEvalCoord1fv", (void **) &__glhw_EvalCoord1fv},
	{"glEvalCoord2d", (void **) &__glhw_EvalCoord2d},
	{"glEvalCoord2dv", (void **) &__glhw_EvalCoord2dv},
	{"glEvalCoord2f", (void **) &__glhw_EvalCoord2f},
	{"glEvalCoord2fv", (void **) &__glhw_EvalCoord2fv},
	{"glEvalMesh1", (void **) &__glhw_EvalMesh1},
	{"glEvalMesh2", (void **) &__glhw_EvalMesh2},
	{"glEvalPoint1", (void **) &__glhw_EvalPoint1},
	{"glEvalPoint2", (void **) &__glhw_EvalPoint2},
	{"glFeedbackBuffer", (void **) &__glhw_FeedbackBuffer},
	{"glFinish", (void **) &__glhw_Finish},
	{"glFlush", (void **) &__glhw_Flush},
	{"glFogf", (void **) &__glhw_Fogf},
	{"glFogfv", (void **) &__glhw_Fogfv},
	{"glFogi", (void **) &__glhw_Fogi},
	{"glFogiv", (void **) &__glhw_Fogiv},
	{"glFrontFace", (void **) &__glhw_FrontFace},
	{"glFrustum", (void **) &__glhw_Frustum},
	{"glGenLists", (void **) &__glhw_GenLists},
	{"glGenTextures", (void **) &__glhw_GenTextures},
	{"glGetBooleanv", (void **) &__glhw_GetBooleanv},
	{"glGetClipPlane", (void **) &__glhw_GetClipPlane},
	{"glGetDoublev", (void **) &__glhw_GetDoublev},
	{"glGetError", (void **) &__glhw_GetError},
	{"glGetFloatv", (void **) &__glhw_GetFloatv},
	{"glGetIntegerv", (void **) &__glhw_GetIntegerv},
	{"glGetLightfv", (void **) &__glhw_GetLightfv},
	{"glGetLightiv", (void **) &__glhw_GetLightiv},
	{"glGetMapdv", (void **) &__glhw_GetMapdv},
	{"glGetMapfv", (void **) &__glhw_GetMapfv},
	{"glGetMapiv", (void **) &__glhw_GetMapiv},
	{"glGetMaterialfv", (void **) &__glhw_GetMaterialfv},
	{"glGetMaterialiv", (void **) &__glhw_GetMaterialiv},
	{"glGetPixelMapfv", (void **) &__glhw_GetPixelMapfv},
	{"glGetPixelMapuiv", (void **) &__glhw_GetPixelMapuiv},
	{"glGetPixelMapusv", (void **) &__glhw_GetPixelMapusv},
	{"glGetPointerv", (void **) &__glhw_GetPointerv},
	{"glGetPolygonStipple", (void **) &__glhw_GetPolygonStipple},
	{"glGetString", (void **) &__glhw_GetString},
	{"glGetTexEnvfv", (void **) &__glhw_GetTexEnvfv},
	{"glGetTexEnviv", (void **) &__glhw_GetTexEnviv},
	{"glGetTexGendv", (void **) &__glhw_GetTexGendv},
	{"glGetTexGenfv", (void **) &__glhw_GetTexGenfv},
	{"glGetTexGeniv", (void **) &__glhw_GetTexGeniv},
	{"glGetTexImage", (void **) &__glhw_GetTexImage},
	{"glGetTexLevelParameterfv", (void **) &__glhw_GetTexLevelParameterfv},
	{"glGetTexLevelParameteriv", (void **) &__glhw_GetTexLevelParameteriv},
	{"glGetTexParameterfv", (void **) &__glhw_GetTexParameterfv},
	{"glGetTexParameteriv", (void **) &__glhw_GetTexParameteriv},
	{"glHint", (void **) &__glhw_Hint},
	{"glIndexMask", (void **) &__glhw_IndexMask},
	{"glIndexPointer", (void **) &__glhw_IndexPointer},
	{"glIndexd", (void **) &__glhw_Indexd},
	{"glIndexdv", (void **) &__glhw_Indexdv},
	{"glIndexf", (void **) &__glhw_Indexf},
	{"glIndexfv", (void **) &__glhw_Indexfv},
	{"glIndexi", (void **) &__glhw_Indexi},
	{"glIndexiv", (void **) &__glhw_Indexiv},
	{"glIndexs", (void **) &__glhw_Indexs},
	{"glIndexsv", (void **) &__glhw_Indexsv},
	{"glIndexub", (void **) &__glhw_Indexub},
	{"glIndexubv", (void **) &__glhw_Indexubv},
	{"glInitNames", (void **) &__glhw_InitNames},
	{"glInterleavedArrays", (void **) &__glhw_InterleavedArrays},
	{"glIsEnabled", (void **) &__glhw_IsEnabled},
	{"glIsList", (void **) &__glhw_IsList},
	{"glIsTexture", (void **) &__glhw_IsTexture},
	{"glLightModelf", (void **) &__glhw_LightModelf},
	{"glLightModelfv", (void **) &__glhw_LightModelfv},
	{"glLightModeli", (void **) &__glhw_LightModeli},
	{"glLightModeliv", (void **) &__glhw_LightModeliv},
	{"glLightf", (void **) &__glhw_Lightf},
	{"glLightfv", (void **) &__glhw_Lightfv},
	{"glLighti", (void **) &__glhw_Lighti},
	{"glLightiv", (void **) &__glhw_Lightiv},
	{"glLineStipple", (void **) &__glhw_LineStipple},
	{"glLineWidth", (void **) &__glhw_LineWidth},
	{"glListBase", (void **) &__glhw_ListBase},
	{"glLoadIdentity", (void **) &__glhw_LoadIdentity},
	{"glLoadMatrixd", (void **) &__glhw_LoadMatrixd},
	{"glLoadMatrixf", (void **) &__glhw_LoadMatrixf},
	{"glLoadName", (void **) &__glhw_LoadName},
	{"glLogicOp", (void **) &__glhw_LogicOp},
	{"glMap1d", (void **) &__glhw_Map1d},
	{"glMap1f", (void **) &__glhw_Map1f},
	{"glMap2d", (void **) &__glhw_Map2d},
	{"glMap2f", (void **) &__glhw_Map2f},
	{"glMapGrid1d", (void **) &__glhw_MapGrid1d},
	{"glMapGrid1f", (void **) &__glhw_MapGrid1f},
	{"glMapGrid2d", (void **) &__glhw_MapGrid2d},
	{"glMapGrid2f", (void **) &__glhw_MapGrid2f},
	{"glMaterialf", (void **) &__glhw_Materialf},
	{"glMaterialfv", (void **) &__glhw_Materialfv},
	{"glMateriali", (void **) &__glhw_Materiali},
	{"glMaterialiv", (void **) &__glhw_Materialiv},
	{"glMatrixMode", (void **) &__glhw_MatrixMode},
	{"glMultMatrixd", (void **) &__glhw_MultMatrixd},
	{"glMultMatrixf", (void **) &__glhw_MultMatrixf},
	{"glNewList", (void **) &__glhw_NewList},
	{"glNormal3b", (void **) &__glhw_Normal3b},
	{"glNormal3bv", (void **) &__glhw_Normal3bv},
	{"glNormal3d", (void **) &__glhw_Normal3d},
	{"glNormal3dv", (void **) &__glhw_Normal3dv},
	{"glNormal3f", (void **) &__glhw_Normal3f},
	{"glNormal3fv", (void **) &__glhw_Normal3fv},
	{"glNormal3i", (void **) &__glhw_Normal3i},
	{"glNormal3iv", (void **) &__glhw_Normal3iv},
	{"glNormal3s", (void **) &__glhw_Normal3s},
	{"glNormal3sv", (void **) &__glhw_Normal3sv},
	{"glNormalPointer", (void **) &__glhw_NormalPointer},
	{"glOrtho", (void **) &__glhw_Ortho},
	{"glPassThrough", (void **) &__glhw_PassThrough},
	{"glPixelMapfv", (void **) &__glhw_PixelMapfv},
	{"glPixelMapuiv", (void **) &__glhw_PixelMapuiv},
	{"glPixelMapusv", (void **) &__glhw_PixelMapusv},
	{"glPixelStoref", (void **) &__glhw_PixelStoref},
	{"glPixelStorei", (void **) &__glhw_PixelStorei},
	{"glPixelTransferf", (void **) &__glhw_PixelTransferf},
	{"glPixelTransferi", (void **) &__glhw_PixelTransferi},
	{"glPixelZoom", (void **) &__glhw_PixelZoom},
	{"glPointSize", (void **) &__glhw_PointSize},
	{"glPolygonMode", (void **) &__glhw_PolygonMode},
	{"glPolygonOffset", (void **) &__glhw_PolygonOffset},
	{"glPolygonStipple", (void **) &__glhw_PolygonStipple},
	{"glPopAttrib", (void **) &__glhw_PopAttrib},
	{"glPopClientAttrib", (void **) &__glhw_PopClientAttrib},
	{"glPopMatrix", (void **) &__glhw_PopMatrix},
	{"glPopName", (void **) &__glhw_PopName},
	{"glPrioritizeTextures", (void **) &__glhw_PrioritizeTextures},
	{"glPushAttrib", (void **) &__glhw_PushAttrib},
	{"glPushClientAttrib", (void **) &__glhw_PushClientAttrib},
	{"glPushMatrix", (void **) &__glhw_PushMatrix},
	{"glPushName", (void **) &__glhw_PushName},
	{"glRasterPos2d", (void **) &__glhw_RasterPos2d},
	{"glRasterPos2dv", (void **) &__glhw_RasterPos2dv},
	{"glRasterPos2f", (void **) &__glhw_RasterPos2f},
	{"glRasterPos2fv", (void **) &__glhw_RasterPos2fv},
	{"glRasterPos2i", (void **) &__glhw_RasterPos2i},
	{"glRasterPos2iv", (void **) &__glhw_RasterPos2iv},
	{"glRasterPos2s", (void **) &__glhw_RasterPos2s},
	{"glRasterPos2sv", (void **) &__glhw_RasterPos2sv},
	{"glRasterPos3d", (void **) &__glhw_RasterPos3d},
	{"glRasterPos3dv", (void **) &__glhw_RasterPos3dv},
	{"glRasterPos3f", (void **) &__glhw_RasterPos3f},
	{"glRasterPos3fv", (void **) &__glhw_RasterPos3fv},
	{"glRasterPos3i", (void **) &__glhw_RasterPos3i},
	{"glRasterPos3iv", (void **) &__glhw_RasterPos3iv},
	{"glRasterPos3s", (void **) &__glhw_RasterPos3s},
	{"glRasterPos3sv", (void **) &__glhw_RasterPos3sv},
	{"glRasterPos4d", (void **) &__glhw_RasterPos4d},
	{"glRasterPos4dv", (void **) &__glhw_RasterPos4dv},
	{"glRasterPos4f", (void **) &__glhw_RasterPos4f},
	{"glRasterPos4fv", (void **) &__glhw_RasterPos4fv},
	{"glRasterPos4i", (void **) &__glhw_RasterPos4i},
	{"glRasterPos4iv", (void **) &__glhw_RasterPos4iv},
	{"glRasterPos4s", (void **) &__glhw_RasterPos4s},
	{"glRasterPos4sv", (void **) &__glhw_RasterPos4sv},
	{"glReadBuffer", (void **) &__glhw_ReadBuffer},
	{"glReadPixels", (void **) &__glhw_ReadPixels},
	{"glRectd", (void **) &__glhw_Rectd},
	{"glRectdv", (void **) &__glhw_Rectdv},
	{"glRectf", (void **) &__glhw_Rectf},
	{"glRectfv", (void **) &__glhw_Rectfv},
	{"glRecti", (void **) &__glhw_Recti},
	{"glRectiv", (void **) &__glhw_Rectiv},
	{"glRects", (void **) &__glhw_Rects},
	{"glRectsv", (void **) &__glhw_Rectsv},
	{"glRenderMode", (void **) &__glhw_RenderMode},
	{"glRotated", (void **) &__glhw_Rotated},
	{"glRotatef", (void **) &__glhw_Rotatef},
	{"glScaled", (void **) &__glhw_Scaled},
	{"glScalef", (void **) &__glhw_Scalef},
	{"glScissor", (void **) &__glhw_Scissor},
	{"glSelectBuffer", (void **) &__glhw_SelectBuffer},
	{"glShadeModel", (void **) &__glhw_ShadeModel},
	{"glStencilFunc", (void **) &__glhw_StencilFunc},
	{"glStencilMask", (void **) &__glhw_StencilMask},
	{"glStencilOp", (void **) &__glhw_StencilOp},
	{"glTexCoord1d", (void **) &__glhw_TexCoord1d},
	{"glTexCoord1dv", (void **) &__glhw_TexCoord1dv},
	{"glTexCoord1f", (void **) &__glhw_TexCoord1f},
	{"glTexCoord1fv", (void **) &__glhw_TexCoord1fv},
	{"glTexCoord1i", (void **) &__glhw_TexCoord1i},
	{"glTexCoord1iv", (void **) &__glhw_TexCoord1iv},
	{"glTexCoord1s", (void **) &__glhw_TexCoord1s},
	{"glTexCoord1sv", (void **) &__glhw_TexCoord1sv},
	{"glTexCoord2d", (void **) &__glhw_TexCoord2d},
	{"glTexCoord2dv", (void **) &__glhw_TexCoord2dv},
	{"glTexCoord2f", (void **) &__glhw_TexCoord2f},
	{"glTexCoord2fv", (void **) &__glhw_TexCoord2fv},
	{"glTexCoord2i", (void **) &__glhw_TexCoord2i},
	{"glTexCoord2iv", (void **) &__glhw_TexCoord2iv},
	{"glTexCoord2s", (void **) &__glhw_TexCoord2s},
	{"glTexCoord2sv", (void **) &__glhw_TexCoord2sv},
	{"glTexCoord3d", (void **) &__glhw_TexCoord3d},
	{"glTexCoord3dv", (void **) &__glhw_TexCoord3dv},
	{"glTexCoord3f", (void **) &__glhw_TexCoord3f},
	{"glTexCoord3fv", (void **) &__glhw_TexCoord3fv},
	{"glTexCoord3i", (void **) &__glhw_TexCoord3i},
	{"glTexCoord3iv", (void **) &__glhw_TexCoord3iv},
	{"glTexCoord3s", (void **) &__glhw_TexCoord3s},
	{"glTexCoord3sv", (void **) &__glhw_TexCoord3sv},
	{"glTexCoord4d", (void **) &__glhw_TexCoord4d},
	{"glTexCoord4dv", (void **) &__glhw_TexCoord4dv},
	{"glTexCoord4f", (void **) &__glhw_TexCoord4f},
	{"glTexCoord4fv", (void **) &__glhw_TexCoord4fv},
	{"glTexCoord4i", (void **) &__glhw_TexCoord4i},
	{"glTexCoord4iv", (void **) &__glhw_TexCoord4iv},
	{"glTexCoord4s", (void **) &__glhw_TexCoord4s},
	{"glTexCoord4sv", (void **) &__glhw_TexCoord4sv},
	{"glTexCoordPointer", (void **) &__glhw_TexCoordPointer},
	{"glTexEnvf", (void **) &__glhw_TexEnvf},
	{"glTexEnvfv", (void **) &__glhw_TexEnvfv},
	{"glTexEnvi", (void **) &__glhw_TexEnvi},
	{"glTexEnviv", (void **) &__glhw_TexEnviv},
	{"glTexGend", (void **) &__glhw_TexGend},
	{"glTexGendv", (void **) &__glhw_TexGendv},
	{"glTexGenf", (void **) &__glhw_TexGenf},
	{"glTexGenfv", (void **) &__glhw_TexGenfv},
	{"glTexGeni", (void **) &__glhw_TexGeni},
	{"glTexGeniv", (void **) &__glhw_TexGeniv},
	{"glTexImage1D", (void **) &__glhw_TexImage1D},
	{"glTexImage2D", (void **) &__glhw_TexImage2D},
	{"glTexParameterf", (void **) &__glhw_TexParameterf},
	{"glTexParameterfv", (void **) &__glhw_TexParameterfv},
	{"glTexParameteri", (void **) &__glhw_TexParameteri},
	{"glTexParameteriv", (void **) &__glhw_TexParameteriv},
	{"glTexSubImage1D", (void **) &__glhw_TexSubImage1D},
	{"glTexSubImage2D", (void **) &__glhw_TexSubImage2D},
	{"glTranslated", (void **) &__glhw_Translated},
	{"glTranslatef", (void **) &__glhw_Translatef},
	{"glVertex2d", (void **) &__glhw_Vertex2d},
	{"glVertex2dv", (void **) &__glhw_Vertex2dv},
	{"glVertex2f", (void **) &__glhw_Vertex2f},
	{"glVertex2fv", (void **) &__glhw_Vertex2fv},
	{"glVertex2i", (void **) &__glhw_Vertex2i},
	{"glVertex2iv", (void **) &__glhw_Vertex2iv},
	{"glVertex2s", (void **) &__glhw_Vertex2s},
	{"glVertex2sv", (void **) &__glhw_Vertex2sv},
	{"glVertex3d", (void **) &__glhw_Vertex3d},
	{"glVertex3dv", (void **) &__glhw_Vertex3dv},
	{"glVertex3f", (void **) &__glhw_Vertex3f},
	{"glVertex3fv", (void **) &__glhw_Vertex3fv},
	{"glVertex3i", (void **) &__glhw_Vertex3i},
	{"glVertex3iv", (void **) &__glhw_Vertex3iv},
	{"glVertex3s", (void **) &__glhw_Vertex3s},
	{"glVertex3sv", (void **) &__glhw_Vertex3sv},
	{"glVertex4d", (void **) &__glhw_Vertex4d},
	{"glVertex4dv", (void **) &__glhw_Vertex4dv},
	{"glVertex4f", (void **) &__glhw_Vertex4f},
	{"glVertex4fv", (void **) &__glhw_Vertex4fv},
	{"glVertex4i", (void **) &__glhw_Vertex4i},
	{"glVertex4iv", (void **) &__glhw_Vertex4iv},
	{"glVertex4s", (void **) &__glhw_Vertex4s},
	{"glVertex4sv", (void **) &__glhw_Vertex4sv},
	{"glVertexPointer", (void **) &__glhw_VertexPointer},
	{"glViewport", (void **) &__glhw_Viewport},
	{"", NULL}
};

/* API for system opengl */
GLapi system_api[] = {
	{ "glAccum", NULL},
	{ "glAlphaFunc", NULL},
	{ "glAreTexturesResident", NULL},
	{ "glArrayElement", NULL},
	{ "glBegin", NULL},
	{ "glBindTexture", NULL},
	{ "glBitmap", NULL},
	{ "glBlendFunc", NULL},
	{ "glCallList", NULL},
	{ "glCallLists", NULL},
	{ "glClear", NULL},
	{ "glClearAccum", NULL},
	{ "glClearColor", NULL},
	{ "glClearDepth", NULL},
	{ "glClearIndex", NULL},
	{ "glClearStencil", NULL},
	{ "glClipPlane", NULL},
	{ "glColor3b", NULL},
	{ "glColor3bv", NULL},
	{ "glColor3d", NULL},
	{ "glColor3dv", NULL},
	{ "glColor3f", NULL},
	{ "glColor3fv", NULL},
	{ "glColor3i", NULL},
	{ "glColor3iv", NULL},
	{ "glColor3s", NULL},
	{ "glColor3sv", NULL},
	{ "glColor3ub", NULL},
	{ "glColor3ubv", NULL},
	{ "glColor3ui", NULL},
	{ "glColor3uiv", NULL},
	{ "glColor3us", NULL},
	{ "glColor3usv", NULL},
	{ "glColor4b", NULL},
	{ "glColor4bv", NULL},
	{ "glColor4d", NULL},
	{ "glColor4dv", NULL},
	{ "glColor4f", NULL},
	{ "glColor4fv", NULL},
	{ "glColor4i", NULL},
	{ "glColor4iv", NULL},
	{ "glColor4s", NULL},
	{ "glColor4sv", NULL},
	{ "glColor4ub", NULL},
	{ "glColor4ubv", NULL},
	{ "glColor4ui", NULL},
	{ "glColor4uiv", NULL},
	{ "glColor4us", NULL},
	{ "glColor4usv", NULL},
	{ "glColorMask", NULL},
	{ "glColorMaterial", NULL},
	{ "glColorPointer", NULL},
	{ "glCopyPixels", NULL},
	{ "glCopyTexImage1D", NULL},
	{ "glCopyTexImage2D", NULL},
	{ "glCopyTexSubImage1D", NULL},
	{ "glCopyTexSubImage2D", NULL},
	{ "glCullFace", NULL},
	{ "glDeleteLists", NULL},
	{ "glDeleteTextures", NULL},
	{ "glDepthFunc", NULL},
	{ "glDepthMask", NULL},
	{ "glDepthRange", NULL},
	{ "glDisable", NULL},
	{ "glDisableClientState", NULL},
	{ "glDrawArrays", NULL},
	{ "glDrawBuffer", NULL},
	{ "glDrawElements", NULL},
	{ "glDrawPixels", NULL},
	{ "glEdgeFlag", NULL},
	{ "glEdgeFlagPointer", NULL},
	{ "glEdgeFlagv", NULL},
	{ "glEnable", NULL},
	{ "glEnableClientState", NULL},
	{ "glEnd", NULL},
	{ "glEndList", NULL},
	{ "glEvalCoord1d", NULL},
	{ "glEvalCoord1dv", NULL},
	{ "glEvalCoord1f", NULL},
	{ "glEvalCoord1fv", NULL},
	{ "glEvalCoord2d", NULL},
	{ "glEvalCoord2dv", NULL},
	{ "glEvalCoord2f", NULL},
	{ "glEvalCoord2fv", NULL},
	{ "glEvalMesh1", NULL},
	{ "glEvalMesh2", NULL},
	{ "glEvalPoint1", NULL},
	{ "glEvalPoint2", NULL},
	{ "glFeedbackBuffer", NULL},
	{ "glFinish", NULL},
	{ "glFlush", NULL},
	{ "glFogf", NULL},
	{ "glFogfv", NULL},
	{ "glFogi", NULL},
	{ "glFogiv", NULL},
	{ "glFrontFace", NULL},
	{ "glFrustum", NULL},
	{ "glGenLists", NULL},
	{ "glGenTextures", NULL},
	{ "glGetBooleanv", NULL},
	{ "glGetClipPlane", NULL},
	{ "glGetDoublev", NULL},
	{ "glGetError", NULL},
	{ "glGetFloatv", NULL},
	{ "glGetIntegerv", NULL},
	{ "glGetLightfv", NULL},
	{ "glGetLightiv", NULL},
	{ "glGetMapdv", NULL},
	{ "glGetMapfv", NULL},
	{ "glGetMapiv", NULL},
	{ "glGetMaterialfv", NULL},
	{ "glGetMaterialiv", NULL},
	{ "glGetPixelMapfv", NULL},
	{ "glGetPixelMapuiv", NULL},
	{ "glGetPixelMapusv", NULL},
	{ "glGetPointerv", NULL},
	{ "glGetPolygonStipple ", NULL},
	{ "glGetString", NULL},
	{ "glGetTexEnvfv", NULL},
	{ "glGetTexEnviv", NULL},
	{ "glGetTexGendv", NULL},
	{ "glGetTexGenfv", NULL},
	{ "glGetTexGeniv", NULL},
	{ "glGetTexImage", NULL},
	{ "glGetTexLevelParameterfv", NULL},
	{ "glGetTexLevelParameteriv", NULL},
	{ "glGetTexParameterfv", NULL},
	{ "glGetTexParameteriv", NULL},
	{ "glHint", NULL},
	{ "glIndexMask", NULL},
	{ "glIndexPointer", NULL},
	{ "glIndexd", NULL},
	{ "glIndexdv", NULL},
	{ "glIndexf", NULL},
	{ "glIndexfv", NULL},
	{ "glIndexi", NULL},
	{ "glIndexiv", NULL},
	{ "glIndexs", NULL},
	{ "glIndexsv", NULL},
	{ "glIndexub", NULL},
	{ "glIndexubv", NULL},
	{ "glInitNames", NULL},
	{ "glInterleavedArrays", NULL},
	{ "glIsEnabled", NULL},
	{ "glIsList", NULL},
	{ "glIsTexture", NULL},
	{ "glLightModelf", NULL},
	{ "glLightModelfv", NULL},
	{ "glLightModeli", NULL},
	{ "glLightModeliv", NULL},
	{ "glLightf", NULL},
	{ "glLightfv", NULL},
	{ "glLighti", NULL},
	{ "glLightiv", NULL},
	{ "glLineStipple", NULL},
	{ "glLineWidth", NULL},
	{ "glListBase", NULL},
	{ "glLoadIdentity", NULL},
	{ "glLoadMatrixd", NULL},
	{ "glLoadMatrixf", NULL},
	{ "glLoadName", NULL},
	{ "glLogicOp", NULL},
	{ "glMap1d", NULL},
	{ "glMap1f", NULL},
	{ "glMap2d", NULL},
	{ "glMap2f", NULL},
	{ "glMapGrid1d", NULL},
	{ "glMapGrid1f", NULL},
	{ "glMapGrid2d", NULL},
	{ "glMapGrid2f", NULL},
	{ "glMaterialf", NULL},
	{ "glMaterialfv", NULL},
	{ "glMateriali", NULL},
	{ "glMaterialiv", NULL},
	{ "glMatrixMode", NULL},
	{ "glMultMatrixd", NULL},
	{ "glMultMatrixf", NULL},
	{ "glNewList", NULL},
	{ "glNormal3b", NULL},
	{ "glNormal3bv", NULL},
	{ "glNormal3d", NULL},
	{ "glNormal3dv", NULL},
	{ "glNormal3f", NULL},
	{ "glNormal3fv", NULL},
	{ "glNormal3i", NULL},
	{ "glNormal3iv", NULL},
	{ "glNormal3s", NULL},
	{ "glNormal3sv", NULL},
	{ "glNormalPointer", NULL},
	{ "glOrtho", NULL},
	{ "glPassThrough", NULL},
	{ "glPixelMapfv", NULL},
	{ "glPixelMapuiv", NULL},
	{ "glPixelMapusv", NULL},
	{ "glPixelStoref", NULL},
	{ "glPixelStorei", NULL},
	{ "glPixelTransferf", NULL},
	{ "glPixelTransferi", NULL},
	{ "glPixelZoom", NULL},
	{ "glPointSize", NULL},
	{ "glPolygonMode", NULL},
	{ "glPolygonOffset", NULL},
	{ "glPolygonStipple", NULL},
	{ "glPopAttrib", NULL},
	{ "glPopClientAttrib", NULL},
	{ "glPopMatrix", NULL},
	{ "glPopName", NULL},
	{ "glPrioritizeTextures", NULL},
	{ "glPushAttrib", NULL},
	{ "glPushClientAttrib", NULL},
	{ "glPushMatrix", NULL},
	{ "glPushName", NULL},
	{ "glRasterPos2d", NULL},
	{ "glRasterPos2dv", NULL},
	{ "glRasterPos2f", NULL},
	{ "glRasterPos2fv", NULL},
	{ "glRasterPos2i", NULL},
	{ "glRasterPos2iv", NULL},
	{ "glRasterPos2s", NULL},
	{ "glRasterPos2sv", NULL},
	{ "glRasterPos3d", NULL},
	{ "glRasterPos3dv", NULL},
	{ "glRasterPos3f", NULL},
	{ "glRasterPos3fv", NULL},
	{ "glRasterPos3i", NULL},
	{ "glRasterPos3iv", NULL},
	{ "glRasterPos3s", NULL},
	{ "glRasterPos3sv", NULL},
	{ "glRasterPos4d", NULL},
	{ "glRasterPos4dv", NULL},
	{ "glRasterPos4f", NULL},
	{ "glRasterPos4fv", NULL},
	{ "glRasterPos4i", NULL},
	{ "glRasterPos4iv", NULL},
	{ "glRasterPos4s", NULL},
	{ "glRasterPos4sv", NULL},
	{ "glReadBuffer", NULL},
	{ "glReadPixels", NULL},
	{ "glRectd", NULL},
	{ "glRectdv", NULL},
	{ "glRectf", NULL},
	{ "glRectfv", NULL},
	{ "glRecti", NULL},
	{ "glRectiv", NULL},
	{ "glRects", NULL},
	{ "glRectsv", NULL},
	{ "glRenderMode", NULL},
	{ "glRotated", NULL},
	{ "glRotatef", NULL},
	{ "glScaled", NULL},
	{ "glScalef", NULL},
	{ "glScissor", NULL},
	{ "glSelectBuffer", NULL},
	{ "glShadeModel", NULL},
	{ "glStencilFunc", NULL},
	{ "glStencilMask", NULL},
	{ "glStencilOp", NULL},
	{ "glTexCoord1d", NULL},
	{ "glTexCoord1dv", NULL},
	{ "glTexCoord1f", NULL},
	{ "glTexCoord1fv", NULL},
	{ "glTexCoord1i", NULL},
	{ "glTexCoord1iv", NULL},
	{ "glTexCoord1s", NULL},
	{ "glTexCoord1sv", NULL},
	{ "glTexCoord2d", NULL},
	{ "glTexCoord2dv", NULL},
	{ "glTexCoord2f", NULL},
	{ "glTexCoord2fv", NULL},
	{ "glTexCoord2i", NULL},
	{ "glTexCoord2iv", NULL},
	{ "glTexCoord2s", NULL},
	{ "glTexCoord2sv", NULL},
	{ "glTexCoord3d", NULL},
	{ "glTexCoord3dv", NULL},
	{ "glTexCoord3f", NULL},
	{ "glTexCoord3fv", NULL},
	{ "glTexCoord3i", NULL},
	{ "glTexCoord3iv", NULL},
	{ "glTexCoord3s", NULL},
	{ "glTexCoord3sv", NULL},
	{ "glTexCoord4d", NULL},
	{ "glTexCoord4dv", NULL},
	{ "glTexCoord4f", NULL},
	{ "glTexCoord4fv", NULL},
	{ "glTexCoord4i", NULL},
	{ "glTexCoord4iv", NULL},
	{ "glTexCoord4s", NULL},
	{ "glTexCoord4sv", NULL},
	{ "glTexCoordPointer", NULL},
	{ "glTexEnvf", NULL},
	{ "glTexEnvfv", NULL},
	{ "glTexEnvi", NULL},
	{ "glTexEnviv", NULL},
	{ "glTexGend", NULL},
	{ "glTexGendv", NULL},
	{ "glTexGenf", NULL},
	{ "glTexGenfv", NULL},
	{ "glTexGeni", NULL},
	{ "glTexGeniv", NULL},
	{ "glTexImage1D", NULL},
	{ "glTexImage2D", NULL},
	{ "glTexParameterf", NULL},
	{ "glTexParameterfv", NULL},
	{ "glTexParameteri", NULL},
	{ "glTexParameteriv", NULL},
	{ "glTexSubImage1D", NULL},
	{ "glTexSubImage2D", NULL},
	{ "glTranslated", NULL},
	{ "glTranslatef", NULL},
	{ "glVertex2d", NULL},
	{ "glVertex2dv", NULL},
	{ "glVertex2f", NULL},
	{ "glVertex2fv", NULL},
	{ "glVertex2i", NULL},
	{ "glVertex2iv", NULL},
	{ "glVertex2s", NULL},
	{ "glVertex2sv", NULL},
	{ "glVertex3d", NULL},
	{ "glVertex3dv", NULL},
	{ "glVertex3f", NULL},
	{ "glVertex3fv", NULL},
	{ "glVertex3i", NULL},
	{ "glVertex3iv", NULL},
	{ "glVertex3s", NULL},
	{ "glVertex3sv", NULL},
	{ "glVertex4d", NULL},
	{ "glVertex4dv", NULL},
	{ "glVertex4f", NULL},
	{ "glVertex4fv", NULL},
	{ "glVertex4i", NULL},
	{ "glVertex4iv", NULL},
	{ "glVertex4s", NULL},
	{ "glVertex4sv", NULL},
	{ "glVertexPointer", NULL},
	{ "glViewport", NULL},
	{ "", NULL}};
	

/* API for default state tracking */
GLapi glcontext_api[] = {
/* glbucket.c */
	{"glAccum", (void *) __glstate_Accum},
	{"glCallList", (void *) __glstate_CallList},
	{"glCallLists", (void *) __glstate_CallLists},
	{"glClear", (void *) __glstate_Clear},
	{"glBitmap", (void *) __glstate_Bitmap},
	{"glDrawPixels", (void *) __glstate_DrawPixels},
	{"glRectd", (void *) __glstate_Rectd},
	{"glRectdv", (void *) __glstate_Rectdv},
	{"glRectf", (void *) __glstate_Rectf},
	{"glRectfv", (void *) __glstate_Rectfv},
	{"glRecti", (void *) __glstate_Recti},
	{"glRectiv", (void *) __glstate_Rectiv},
	{"glRects", (void *) __glstate_Rects},
	{"glRectsv", (void *) __glstate_Rectsv},
/* glbuffer.c */
	{ "glAlphaFunc", (void *) __glstate_AlphaFunc},
	{ "glDepthFunc", (void *) __glstate_DepthFunc},
	{ "glBlendFunc", (void *) __glstate_BlendFunc},
	{ "glLogicOp", (void *) __glstate_LogicOp},
	{ "glDrawBuffer", (void *) __glstate_DrawBuffer},
	{ "glReadBuffer", (void *) __glstate_ReadBuffer},
	{ "glIndexMask", (void *) __glstate_IndexMask},
	{ "glColorMask", (void *) __glstate_ColorMask},
	{ "glClearColor", (void *) __glstate_ClearColor},
	{ "glClearIndex", (void *) __glstate_ClearIndex},
	{ "glClearDepth", (void *) __glstate_ClearDepth},
	{ "glClearAccum", (void *) __glstate_ClearAccum},
	{ "glDepthMask", (void *) __glstate_DepthMask},
/* glclient.c */
	{ "glEnableClientState", (void *) __glstate_EnableClientState},
	{ "glDisableClientState", (void *) __glstate_DisableClientState},
	{ "glVertexPointer", (void *) __glstate_VertexPointer},
	{ "glColorPointer", (void *) __glstate_ColorPointer},
	{ "glIndexPointer", (void *) __glstate_IndexPointer},
	{ "glNormalPointer", (void *) __glstate_NormalPointer},
	{ "glTexCoordPointer", (void *) __glstate_TexCoordPointer},
	{ "glEdgeFlagPointer", (void *) __glstate_EdgeFlagPointer},
	{ "glInterleavedArrays", (void *) __glstate_InterleavedArrays},
	{ "glDrawArrays", (void *) __glstate_DrawArrays}, 
	{ "glArrayElement", (void *) __glstate_ArrayElement}, 
	{ "glGetPointerv", (void *) __glstate_GetPointerv},
	{ "glDrawArrays", (void *) __glstate_DrawArrays},
	{ "glDrawElements", (void *) __glstate_DrawElements},
/* glcolor.c */
	{"glColor3b", (void *) __glstate_Color3b},
	{"glColor3bv", (void *) __glstate_Color3bv},
	{"glColor3d", (void *) __glstate_Color3d},
	{"glColor3dv", (void *) __glstate_Color3dv},
	{"glColor3f", (void *) __glstate_Color3f},
	{"glColor3fv", (void *) __glstate_Color3fv},
	{"glColor3i", (void *) __glstate_Color3i},
	{"glColor3iv", (void *) __glstate_Color3iv},
	{"glColor3s", (void *) __glstate_Color3s},
	{"glColor3sv", (void *) __glstate_Color3sv},
	{"glColor3ub", (void *) __glstate_Color3ub},
	{"glColor3ubv", (void *) __glstate_Color3ubv},
	{"glColor3ui", (void *) __glstate_Color3ui},
	{"glColor3uiv", (void *) __glstate_Color3uiv},
	{"glColor3us", (void *) __glstate_Color3us},
	{"glColor3usv", (void *) __glstate_Color3usv},
	{"glColor4b", (void *) __glstate_Color4b},
	{"glColor4bv", (void *) __glstate_Color4bv},
	{"glColor4d", (void *) __glstate_Color4d},
	{"glColor4dv", (void *) __glstate_Color4dv},
	{"glColor4f", (void *) __glstate_Color4f},
	{"glColor4fv", (void *) __glstate_Color4fv},
	{"glColor4i", (void *) __glstate_Color4i},
	{"glColor4iv", (void *) __glstate_Color4iv},
	{"glColor4s", (void *) __glstate_Color4s},
	{"glColor4sv", (void *) __glstate_Color4sv},
	{"glColor4ub", (void *) __glstate_Color4ub},
	{"glColor4ubv", (void *) __glstate_Color4ubv},
	{"glColor4ui", (void *) __glstate_Color4ui},
	{"glColor4uiv", (void *) __glstate_Color4uiv},
	{"glColor4us", (void *) __glstate_Color4us},
	{"glColor4usv", (void *) __glstate_Color4usv},
/* glbucket.c */
	{"glBegin", (void *) __glstate_Begin},
	{"glEnd", (void *) __glstate_End},
/* glenable.c */
	{ "glEnable", (void *) __glstate_Enable},
	{ "glDisable", (void *) __glstate_Disable},
/* gleval.c */
	{"glMap1d", (void *) __glstate_Map1d},
	{"glMap1f", (void *) __glstate_Map1f},
	{"glMap2d", (void *) __glstate_Map2d},
	{"glMap2f", (void *) __glstate_Map2f},
	{"glMapGrid1d", (void *) __glstate_MapGrid1d},
	{"glMapGrid1f", (void *) __glstate_MapGrid1f},
	{"glMapGrid2d", (void *) __glstate_MapGrid2d},
	{"glMapGrid2f", (void *) __glstate_MapGrid2f},
/* glfog.c */
	{ "glFogf", (void *) __glstate_Fogf},
	{ "glFogi", (void *) __glstate_Fogi},
	{ "glFogiv", (void *) __glstate_Fogiv},
	{ "glFogfv", (void *) __glstate_Fogfv},
/* glget.c */
	{"glGetBooleanv", (void *) __glstate_GetBooleanv},
	{"glGetIntegerv", (void *) __glstate_GetIntegerv},
	{"glGetFloatv", (void *) __glstate_GetFloatv},
	{"glGetDoublev", (void *) __glstate_GetDoublev},
	{"glIsEnabled", (void *) __glstate_IsEnabled},
	{"glGetString", (void *) __glstate_GetString},
/* glhint.c */
	{ "glHint", (void *) __glstate_Hint},
/* gllighting.c */
	{ "glShadeModel", (void *) __glstate_ShadeModel},
	{ "glColorMaterial", (void *) __glstate_ColorMaterial},
	{ "glLightModelf", (void *) __glstate_LightModelf},
	{ "glLightModeli", (void *) __glstate_LightModeli},
	{ "glLightModelfv", (void *) __glstate_LightModelfv},
	{ "glLightModeliv", (void *) __glstate_LightModeliv},
	{ "glLightf", (void *) __glstate_Lightf},
	{ "glLighti", (void *) __glstate_Lighti},
	{ "glLightfv", (void *) __glstate_Lightfv},
	{ "glLightiv", (void *) __glstate_Lightiv},
	{ "glMaterialf", (void *) __glstate_Materialf},
	{ "glMateriali", (void *) __glstate_Materiali},
	{ "glMaterialfv", (void *) __glstate_Materialfv},
	{ "glMaterialiv", (void *) __glstate_Materialiv},
	{"glGetMaterialfv", (void *) __glstate_GetMaterialfv},
	{"glGetMaterialiv", (void *) __glstate_GetMaterialiv},
	{"glGetLightfv", (void *) __glstate_GetLightfv},
	{"glGetLightiv", (void *) __glstate_GetLightiv},
/* glline.c */
	{ "glPointSize", (void *) __glstate_PointSize},
	{ "glLineWidth", (void *) __glstate_LineWidth},
	{ "glLineStipple", (void *) __glstate_LineStipple},
/* gllists.c */
	{"glGenLists", (void *) __glstate_GenLists},
	{"glNewList", (void *) __glstate_NewList},
	{"glEndList", (void *) __glstate_EndList},
	{"glDeleteLists", (void *) __glstate_DeleteLists},
	{"glIsList", (void *) __glstate_IsList},
	{"glListBase", (void *) __glstate_ListBase},
/* glpixel.c */
	{ "glPixelStoref", (void *) __glstate_PixelStoref},
	{ "glPixelStorei", (void *) __glstate_PixelStorei},
	{ "glPixelTransferf", (void *) __glstate_PixelTransferf},
	{ "glPixelTransferi", (void *) __glstate_PixelTransferi},
	{ "glPixelZoom", (void *) __glstate_PixelZoom},
/* glpolygon.c */
	{ "glCullFace", (void *) __glstate_CullFace},
	{ "glFrontFace", (void *) __glstate_FrontFace},
	{ "glPolygonMode", (void *) __glstate_PolygonMode},
	{ "glPolygonOffset", (void *) __glstate_PolygonOffset},
	{ "glPolygonStipple", (void *) __glstate_PolygonStipple},
	{ "glGetPolygonStipple", (void *) __glstate_GetPolygonStipple},
/* glrasterpos.c */
	{"glRasterPos2d", (void *) __glstate_RasterPos2d},
	{"glRasterPos2dv", (void *) __glstate_RasterPos2dv},
	{"glRasterPos2f", (void *) __glstate_RasterPos2f},
	{"glRasterPos2fv", (void *) __glstate_RasterPos2fv},
	{"glRasterPos2i", (void *) __glstate_RasterPos2i},
	{"glRasterPos2iv", (void *) __glstate_RasterPos2iv},
	{"glRasterPos2s", (void *) __glstate_RasterPos2s},
	{"glRasterPos2sv", (void *) __glstate_RasterPos2sv},
	{"glRasterPos3d", (void *) __glstate_RasterPos3d},
	{"glRasterPos3dv", (void *) __glstate_RasterPos3dv},
	{"glRasterPos3f", (void *) __glstate_RasterPos3f},
	{"glRasterPos3fv", (void *) __glstate_RasterPos3fv},
	{"glRasterPos3i", (void *) __glstate_RasterPos3i},
	{"glRasterPos3iv", (void *) __glstate_RasterPos3iv},
	{"glRasterPos3s", (void *) __glstate_RasterPos3s},
	{"glRasterPos3sv", (void *) __glstate_RasterPos3sv},
	{"glRasterPos4d", (void *) __glstate_RasterPos4d},
	{"glRasterPos4dv", (void *) __glstate_RasterPos4dv},
	{"glRasterPos4f", (void *) __glstate_RasterPos4f},
	{"glRasterPos4fv", (void *) __glstate_RasterPos4fv},
	{"glRasterPos4i", (void *) __glstate_RasterPos4i},
	{"glRasterPos4iv", (void *) __glstate_RasterPos4iv},
	{"glRasterPos4s", (void *) __glstate_RasterPos4s},
	{"glRasterPos4sv", (void *) __glstate_RasterPos4sv},
/* glscissor.c */
	{ "glScissor", (void *) __glstate_Scissor},
/* glstencil.c */
	{ "glStencilFunc", (void *) __glstate_StencilFunc},
	{ "glStencilOp", (void *) __glstate_StencilOp},
	{ "glClearStencil", (void *) __glstate_ClearStencil},
	{ "glStencilMask", (void *) __glstate_StencilMask},
/* gltexture.c */
	{ "glGenTextures", (void *) __glstate_GenTextures},
	{ "glDeleteTextures", (void *) __glstate_DeleteTextures},
	{ "glBindTexture", (void *) __glstate_BindTexture},
	{ "glTexImage1D", (void *) __glstate_TexImage1D},
	{ "glTexImage2D", (void *) __glstate_TexImage2D},
	{ "glTexSubImage1D", (void *) __glstate_TexSubImage1D},
	{ "glTexSubImage2D", (void *) __glstate_TexSubImage2D},
	{ "glTexImage3D", (void *) __glstate_TexImage3D},
	{ "glTexParameteri", (void *) __glstate_TexParameteri},
	{ "glTexParameterf", (void *) __glstate_TexParameterf},
	{ "glTexParameteriv", (void *) __glstate_TexParameteriv},
	{ "glTexParameterfv", (void *) __glstate_TexParameterfv},
	{ "glGetTexParameteriv", (void *) __glstate_GetTexParameteriv},
	{ "glGetTexParameterfv", (void *) __glstate_GetTexParameterfv},
	{ "glGetTexLevelParameterfv", (void *) __glstate_GetTexLevelParameterfv},
	{ "glGetTexLevelParameteriv", (void *) __glstate_GetTexLevelParameteriv},
	{ "glTexEnvf", (void *) __glstate_TexEnvf},
	{ "glTexEnvi", (void *) __glstate_TexEnvi},
	{ "glTexEnvfv", (void *) __glstate_TexEnvfv},
	{ "glTexEnviv", (void *) __glstate_TexEnviv},
	{ "glGetTexEnvfv", (void *) __glstate_GetTexEnvfv},
	{ "glGetTexEnviv", (void *) __glstate_GetTexEnviv},
	{ "glTexGend", (void *) __glstate_TexGend},
	{ "glTexGenf", (void *) __glstate_TexGenf},
	{ "glTexGeni", (void *) __glstate_TexGeni},
	{ "glTexGendv", (void *) __glstate_TexGendv},
	{ "glTexGenfv", (void *) __glstate_TexGenfv},
	{ "glTexGeniv", (void *) __glstate_TexGeniv},
	{ "glGetTexGendv", (void *) __glstate_GetTexGendv},
	{ "glGetTexGenfv", (void *) __glstate_GetTexGenfv},
	{ "glGetTexGeniv", (void *) __glstate_GetTexGeniv},
	{ "glAreTexturesResident", (void *) __glstate_AreTexturesResident},
	{ "glPrioritizeTextures", (void *) __glstate_PrioritizeTextures},
/* gltrans.c */
	{ "glClipPlane", (void *) __glstate_ClipPlane},
	{ "glGetClipPlane", (void *) __glstate_GetClipPlane},
	{ "glMatrixMode", (void *) __glstate_MatrixMode},
	{ "glLoadIdentity", (void *) __glstate_LoadIdentity},
	{ "glPopMatrix", (void *) __glstate_PopMatrix},
	{ "glPushMatrix", (void *) __glstate_PushMatrix},
	{ "glLoadMatrixf", (void *) __glstate_LoadMatrixf},
	{ "glLoadMatrixd", (void *) __glstate_LoadMatrixd},
	{ "glMultMatrixf", (void *) __glstate_MultMatrixf},
	{ "glMultMatrixd", (void *) __glstate_MultMatrixd},
	{ "glTranslatef", (void *) __glstate_Translatef},
	{ "glTranslated", (void *) __glstate_Translated},
	{ "glRotatef", (void *) __glstate_Rotatef},
	{ "glRotated", (void *) __glstate_Rotated},
	{ "glScalef", (void *) __glstate_Scalef},
	{ "glScaled", (void *) __glstate_Scaled},
	{ "glFrustum", (void *) __glstate_Frustum},
	{ "glOrtho", (void *) __glstate_Ortho},
/* glviewport.c */
	{ "glViewport", (void *) __glstate_Viewport},
	{ "glDepthRange", (void *) __glstate_DepthRange},
	{ "glGetError", (void *) __glstate_GetError},
	{ "", NULL}};

void GLAPI 
__glhw_nop() {}

void GLAPI
__glhw_nop3 (GLbyte, GLbyte, GLbyte) {}

void GLAPI
__glhw_nop4 (GLint) {}

void GLAPI
__glhw_nop6 (GLshort, GLshort, GLshort) {}

void GLAPI
__glhw_nop8 (GLdouble) {}

void GLAPI
__glhw_nop12 (GLdouble, GLint) {}

void GLAPI
__glhw_nop16 (GLdouble, GLdouble) {}

void GLAPI
__glhw_nop20 (GLdouble, GLdouble, GLint) {}

void GLAPI
__glhw_nop24 (GLdouble, GLdouble, GLdouble) {}

void GLAPI
__glhw_nop28 (GLdouble, GLdouble, GLdouble, GLint) {}

void GLAPI
__glhw_nop32 (GLdouble, GLdouble, GLdouble, GLdouble) {}

void GLAPI
__glhw_nop36 (GLdouble, GLdouble, GLdouble, GLdouble, GLint) {}

void GLAPI
__glhw_nop40 (GLdouble, GLdouble, GLdouble, GLdouble, GLdouble) {}

void GLAPI
__glhw_nop44 (GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLint) {}

void GLAPI
__glhw_nop48 (GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble) {}

void GLAPI
__glhw_nop52 (GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLint) {}

void GLAPI
__glhw_nop56 (GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble, GLdouble) {}


/* NOP assignments */
GLapi nop_api[] = {
/* glbuffer.c */
	{ "glAlphaFunc", (void *) __glhw_nop8},
	{ "glDepthFunc", (void *) __glhw_nop4},
	{ "glBlendFunc", (void *) __glhw_nop8},
	{ "glBlendColor", (void *) __glhw_nop16},
	{ "glBlendEquation", (void *) __glhw_nop4},
	{ "glLogicOp", (void *) __glhw_nop4},
	{ "glDrawBuffer", (void *) __glhw_nop4},
	{ "glReadBuffer", (void *) __glhw_nop4},
	{ "glIndexMask", (void *) __glhw_nop4},
	{ "glColorMask", (void *) __glhw_nop4},
	{ "glClearColor", (void *) __glhw_nop16},
	{ "glClearIndex", (void *) __glhw_nop4},
	{ "glClearDepth", (void *) __glhw_nop4},
	{ "glClearAccum", (void *) __glhw_nop16},
	{ "glDepthMask", (void *) __glhw_nop4},
/* glclient.c */
	{ "glEnableClientState", (void *) __glhw_nop4},
	{ "glDisableClientState", (void *) __glhw_nop4},
	{ "glVertexPointer", (void *) __glhw_nop16},
	{ "glColorPointer", (void *) __glhw_nop16},
	{ "glIndexPointer", (void *) __glhw_nop12},
	{ "glNormalPointer", (void *) __glhw_nop12},
	{ "glTexCoordPointer", (void *) __glhw_nop16},
	{ "glEdgeFlagPointer", (void *) __glhw_nop8},
/* glcurrent.c */
	{ "glColor3bv", (void *) __glhw_nop3},
	{ "glColor3dv", (void *) __glhw_nop24},
	{ "glColor3fv", (void *) __glhw_nop12},
	{ "glColor3iv", (void *) __glhw_nop12},
	{ "glColor3sv", (void *) __glhw_nop6},
	{ "glColor3ubv", (void *) __glhw_nop3},
	{ "glColor3uiv", (void *) __glhw_nop12},
	{ "glColor3usv", (void *) __glhw_nop6},
	{ "glColor4bv", (void *) __glhw_nop4},
	{ "glColor4dv", (void *) __glhw_nop32},
	{ "glColor4fv", (void *) __glhw_nop16},
	{ "glColor4iv", (void *) __glhw_nop16},
	{ "glColor4sv", (void *) __glhw_nop8},
	{ "glColor4ubv", (void *) __glhw_nop4},
	{ "glColor4uiv", (void *) __glhw_nop16},
	{ "glColor4usv", (void *) __glhw_nop8},
	{ "glBegin", (void *) __glhw_nop4},
	{ "glEnd", (void *) __glhw_nop},
/* glenable.c */
	{ "glEnable", (void *) __glhw_nop4},
	{ "glDisable", (void *) __glhw_nop4},
	{ "glDepthMask", (void *) __glhw_nop4},
/* gleval.c */
	{ "glMap1d", (void *) __glhw_nop32},
	{ "glMap2d", (void *) __glhw_nop56},
	{ "glMapGrid1d", (void *) __glhw_nop20},
	{ "glMapGrid2d", (void *) __glhw_nop40},
/* glfog.c */
	{ "glFogf", (void *) __glhw_nop8},
	{ "glFogi", (void *) __glhw_nop8},
	{ "glFogiv", (void *) __glhw_nop8},
	{ "glFogfv", (void *) __glhw_nop8},
/* glhint.c */
	{ "glHint", (void *) __glhw_nop8},
/* gllighting.c */
	{ "glShadeModel", (void *) __glhw_nop4},
	{ "glColorMaterial", (void *) __glhw_nop8},
	{ "glLightModelf", (void *) __glhw_nop8},
	{ "glLightModeli", (void *) __glhw_nop8},
	{ "glLightModelfv", (void *) __glhw_nop8},
	{ "glLightModeliv", (void *) __glhw_nop8},
	{ "glLighf", (void *) __glhw_nop12},
	{ "glLighti", (void *) __glhw_nop12},
	{ "glLightfv", (void *) __glhw_nop12},
	{ "glLightiv", (void *) __glhw_nop12},
	{ "glMaterialf", (void *) __glhw_nop12},
	{ "glMateriali", (void *) __glhw_nop12},
	{ "glMaterialfv", (void *) __glhw_nop12},
	{ "glMaterialiv", (void *) __glhw_nop12},
/* glline.c */
	{ "glPointSize", (void *) __glhw_nop4},
	{ "glLineWidth", (void *) __glhw_nop4},
	{ "glLineStipple", (void *) __glhw_nop8},
/* glpixel.c */
	{ "glPixelStoref", (void *) __glhw_nop8},
	{ "glPixelStorei", (void *) __glhw_nop8}, 
	{ "glPixelTransferf", (void *) __glhw_nop8},
	{ "glPixelTransferi", (void *) __glhw_nop8},
	{ "glPixelZoom", (void *) __glhw_nop8},
/* glpolygon.c */
	{ "glCullFace", (void *) __glhw_nop4},
	{ "glFrontFace", (void *) __glhw_nop4},
	{ "glPolygonMode", (void *) __glhw_nop8},
	{ "glPolygonStipple", (void *) __glhw_nop4}, 
/* glscissor.c */
	{ "glScissor", (void *) __glhw_nop16},
/* glstencil.c */
	{ "glStencilFunc", (void *) __glhw_nop12},
	{ "glStencilOp", (void *) __glhw_nop12},
	{ "glClearStencil", (void *) __glhw_nop4},
	{ "glStencilMask", (void *) __glhw_nop4},
/* gltexture.c */
	{ "glBindTexture", (void *) __glhw_nop8},
	{ "glTexParameteri", (void *) __glhw_nop12},
	{ "glTexParameterf", (void *) __glhw_nop12},
	{ "glTexParameteriv", (void *) __glhw_nop12},
	{ "glTexParameterfv", (void *) __glhw_nop12},
	{ "glTexImage1D", (void *) __glhw_nop32},
	{ "glTexImage2D", (void *) __glhw_nop36},
	{ "glTexImage3D", (void *) __glhw_nop40},
	{ "glTexGend", (void *) __glhw_nop16},
	{ "glTexGenf", (void *) __glhw_nop12},
	{ "glTexGeni", (void *) __glhw_nop12},
	{ "glTexGendv", (void *) __glhw_nop12},
	{ "glTexGenfv", (void *) __glhw_nop12},
	{ "glTexEnvf", (void *) __glhw_nop12},
	{ "glTexEnvi", (void *) __glhw_nop12},
	{ "glTexEnvfv", (void *) __glhw_nop12},
	{ "glTexEnviv", (void *) __glhw_nop12},
/* gltrans.c */
	{ "glClipPlane", (void *) __glhw_nop8},
	{ "glGetClipPlane", (void *) __glhw_nop8},
	{ "glMatrixMode", (void *) __glhw_nop},
	{ "glLoadIdentity", (void *) __glhw_nop},
	{ "glPopMatrix", (void *) __glhw_nop},
	{ "glPushMatrix", (void *) __glhw_nop},
	{ "glLoadMatrixf", (void *) __glhw_nop4},
	{ "glLoadMatrixd", (void *) __glhw_nop4},
	{ "glMultMatrixf", (void *) __glhw_nop4},
	{ "glMultMatrixd", (void *) __glhw_nop4},
	{ "glTranslatef", (void *) __glhw_nop12},
	{ "glTranslated", (void *) __glhw_nop24},
	{ "glRotatef", (void *) __glhw_nop16},
	{ "glRotated", (void *) __glhw_nop32},
	{ "glScalef", (void *) __glhw_nop12},
	{ "glScaled", (void *) __glhw_nop24},
	{ "glFrustum", (void *) __glhw_nop48},
	{ "glOrtho", (void *) __glhw_nop48},
/* glviewport.c */
	{ "glViewport", (void *) __glhw_nop16},
	{ "glDepthRange", (void *) __glhw_nop16},
/* glconfig.c */
	{ "glGetBooleanv", (void *) __glhw_nop8}, 
	{ "glGetIntegerv", (void *) __glhw_nop8}, 
	{ "glGetFloatv", (void *) __glhw_nop8},
	{ "glGetConvolutionParameteriv", (void *) __glhw_nop12},
	{ "glGetString", (void *) __glhw_nop4},
	{ "", NULL}
};

