/*
** wiregl_util.h
** http://graphics.stanford.edu/software/wiregl
**
** Copyright  2000 The Board of Trustees of The Leland Stanford Junior
** University. All rights reserved.
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or modify
** this software, provided that the above copyright notice and this
** permission notice appear in all copies of this software.  No part of
** this software or any derivatives thereof may be used in graphics systems
** for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind, 
** express, implied or otherwise, including without limitation, any warranty 
** of merchantability or fitness for a particular purpose.
**
*/

#ifndef WIREGL_UTIL_H
#define WIREGL_UTIL_H

#ifndef WINDOWS
#define UTIL_DECL
#else
#ifndef UTIL_DECL
#define UTIL_DECL __declspec(dllimport)
#endif
#endif

#ifdef __GNUC__
#define NORETURN __attribute__ ((noreturn))
#else
#define NORETURN
#endif

#if defined( IRIX ) || defined( IRIX64 )
#include <ulocks.h>

extern usptr_t *shared_mem;
#endif

#if defined (WINDOWS)
#include <windows.h>
#include <winbase.h>
#include <winsock.h>
#endif

#if defined(Linux)
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <string.h>
#include <netinet/in.h>
#include <unistd.h>
#include <semaphore.h>
#include <pthread.h>
#endif	/*Linux*/

#include <stdio.h> // for FILE *

#include "wiregl/include/wiregl_common.h"
 
#ifdef __cplusplus
extern "C" {
#endif

UTIL_DECL void *wireGLAlloc( unsigned int bytes );
UTIL_DECL void wireGLFree( void *ptr );

typedef struct {
#if defined( IRIX ) || defined( IRIX64 )
	usema_t *sema;
#endif
#if defined (WINDOWS)
  HANDLE sema;
#endif
#if defined (Linux)
    sem_t sema;
#endif
} WireGLSemaphore;

#if defined(WINDOWS)
typedef volatile LONG WireGLSpinLock;
#elif defined(Linux)
typedef volatile long WireGLSpinLock;
#else
#error SPINLOCKS
#endif

typedef struct {
	volatile int *someone_waiting;
	WireGLSemaphore *threadSema;
} WireGLCondVar;

extern WireGLSpinLock output_sema;
UTIL_DECL void wireGLInitSemaphore( WireGLSemaphore *m, int value, char *name );
UTIL_DECL void wireGLAcquireSemaphore( WireGLSemaphore *m );
UTIL_DECL void wireGLReleaseSemaphore( WireGLSemaphore *m );
UTIL_DECL void wireGLDestroySemaphore( WireGLSemaphore *m );

UTIL_DECL void wireGLInitSpinLock( WireGLSpinLock *sl );
UTIL_DECL void wireGLAcquireSpinLock( WireGLSpinLock *sl );
UTIL_DECL void wireGLReleaseSpinLock( WireGLSpinLock *sl );

UTIL_DECL void wireGLInitCondVar( WireGLCondVar *cv, WireGLSemaphore *threadSema, volatile int *flag );
UTIL_DECL void wireGLWaitCondVar( WireGLCondVar *cv, WireGLSpinLock *sl );
UTIL_DECL void wireGLSignalCondVar( WireGLCondVar *cv );

typedef enum {
  WIREGL_WARN_NONE=0,
  WIREGL_WARN_CRITICAL,
  WIREGL_WARN_IMPORTANT,
  WIREGL_WARN_NOTICE,
  WIREGL_WARN_CHATTY,
  WIREGL_WARN_DEBUG,
  WIREGL_WARN_PACKET_DEBUG,
  WIREGL_WARN_VERBOSE_DEBUG
} WireGLWarningLevel;

#ifdef __cplusplus
}
#endif

UTIL_DECL void wireGLError( const char *format, ... ) NORETURN;
UTIL_DECL void wireGLSimpleError( const char *format, ... ) NORETURN;
UTIL_DECL void wireGLPrintError( const char *format, ... ) NORETURN;
UTIL_DECL void wireGLWarning( WireGLWarningLevel level, const char *format, ... );
UTIL_DECL void wireGLInitWarnings( void );
UTIL_DECL char *wireGLInitOS(void);
UTIL_DECL char wireGLDetermineEndianness();
UTIL_DECL void wireGLSleep( unsigned int msecs );

struct _WireGLConnection;

typedef int (*WireGLReader)( struct _WireGLConnection *, void * );
typedef void (*WireGLSender)( struct _WireGLConnection *, void *, int );
typedef void (*WireGLDMARegistrar)( struct _WireGLConnection *, void *, int );

typedef enum {
	TCPIP,
	SHARED_MEM,
	DROP_PACKETS,
	FILE_TRACE,
	GM
} WireGLConnectionType;

typedef struct {
	int consumer_head_offset;
	int consumer_tail_offset;
	int producer_head_offset;
	int producer_tail_offset;
	WireGLSpinLock producer_mutex;
	WireGLSpinLock consumer_mutex;

	int bytes_produced, bytes_consumed;

	volatile int waiting_flag;
} WireGLSharedMemoryData;

typedef struct _WireGLConnection {
	WireGLConnectionType type;

	char *hostname;
	char please_share_memory, please_use_gm;

	char *receive_buf_one, *receive_buf_two;

	WireGLReader wireGLReadMessage;
	WireGLSender wireGLSendMessage;
	WireGLSender wireGLInjectMessage;
	WireGLDMARegistrar wireGLRegisterMemoryForDMA;
	WireGLDMARegistrar wireGLDeregisterMemoryForDMA;

	// asyncronous data
	void *data_to_send;
	int data_length;
	void *data_to_send2;
	int data_length2;
	WireGLSemaphore data_ready;
	WireGLSemaphore data_sent;
	int use_async;

	// TCP/IP
	int tcp_socket;

	// Shared Memory
	WireGLSharedMemoryData *shmem;
	WireGLSemaphore threadSema;
	char *map_filename;
	char *consumer_start, *consumer_end;
	char *producer_start, *producer_end;
	int * volatile p_consumer_head_offset;
	int * volatile p_consumer_tail_offset;
	int * volatile p_producer_head_offset;
	int * volatile p_producer_tail_offset;
	WireGLCondVar *bytesConsumed, *bytesProduced;
	WireGLSpinLock *producer_lock, *consumer_lock;

	// File trace
	FILE *fp;
	char *file_start, *file_buf, *file_end;

	//GM
	struct gm_port *wiregl_gm_port;
	unsigned int wiregl_gm_port_number;
	unsigned int wiregl_gm_server_id;

	int wiregl_gm_spinning_for_fun;

	char *wiregl_gm_zero_length_hack;

} WireGLConnection;

UTIL_DECL void wireGLCloseSocket( int sock );

UTIL_DECL void wireGLSendAsync( WireGLConnection *conn, void *data, int length );
UTIL_DECL void wireGLSendTwoAsync( WireGLConnection *conn, void *data, int length, void *data2, int length2 );

typedef struct {
	int sock;
	struct sockaddr_in addr;
} WireGLServer;

UTIL_DECL WireGLServer *wireGLBecomeServer( int port );
UTIL_DECL WireGLConnection *wireGLConnectToServer( char *server, int default_port, WireGLClientConnection *connection );
UTIL_DECL WireGLConnection *wireGLAcceptClient( WireGLServer *server, WireGLClientConnection *client_connection );
UTIL_DECL WireGLConnection *wireGLAcceptFileClient( char *fname );

UTIL_DECL void wireGLMakeShmem( WireGLConnection *conn );
UTIL_DECL void wireGLSetUpShmem( WireGLConnection *conn, char *map_filename );
UTIL_DECL void wireGLStartShmem( WireGLConnection *conn );

UTIL_DECL void wireGLFinishConnectToServer( WireGLConnection * );
UTIL_DECL void wireGLFinishAcceptClient( WireGLConnection * );

UTIL_DECL void wireGLMakeGM( WireGLConnection *conn );
UTIL_DECL void wireGLStartGM( WireGLConnection *conn );

#define WIREGL_MAX_PACKET_SIZE 16*1024*1024

typedef void (*WireGLThreadFunc)( void * );
UTIL_DECL void wireGLCreateThread( WireGLThreadFunc entry, void *arg );

#define WIREGL_NUM_BUCKETS 1047

typedef struct _wireglhashnode {
	char is_string;
	unsigned int key;
	char *str_key;
	void *data;
	struct _wireglhashnode *next;
} WireGLHashNode;

typedef struct {
	unsigned int num_elements;
	WireGLHashNode *buckets[WIREGL_NUM_BUCKETS];
	WireGLSpinLock mutex;
} WireGLHashTable;

UTIL_DECL WireGLHashTable *wireGLAllocHashtable( void );
UTIL_DECL void wireGLFreeHashtable( WireGLHashTable *hash );
UTIL_DECL void wireGLHashtableAdd( WireGLHashTable *h, unsigned int key, void *data );
UTIL_DECL void wireGLHashtableStringAdd( WireGLHashTable *h, char *key, void *data );
UTIL_DECL void wireGLHashtableDelete( WireGLHashTable *h, unsigned int key );
UTIL_DECL void wireGLHashtableStringDelete( WireGLHashTable *h, char *key );
UTIL_DECL void *wireGLHashtableSearch( WireGLHashTable *h, unsigned int key );
UTIL_DECL void *wireGLHashtableStringSearch( WireGLHashTable *h, char * key );
UTIL_DECL void wireGLHashtableReplace( WireGLHashTable *h, unsigned int key, void *data, int free_mem );
UTIL_DECL void wireGLHashtableStringReplace( WireGLHashTable *h, char * key, void *data, int free_mem );
UTIL_DECL unsigned int wireGLHashtableNumElements( WireGLHashTable *h) ;

#define WIREGL_HASHTABLE_WALK( h, t ) {         \
  WireGLHashNode *t;                            \
  int _;                                        \
  wireGLAcquireSpinLock( &( h->mutex) );        \
  for (_ = 0 ; _ < WIREGL_NUM_BUCKETS ; _++) {  \
    for (t = h->buckets[_] ; t; t = t->next)   {


#define WIREGL_HASHTABLE_WALK_END( h )          \
  }                                             \
 }                                              \
 wireGLReleaseSpinLock( &( h->mutex ) );        \
}


#ifndef WINDOWS
typedef unsigned long long iotimer64_t;
typedef unsigned int iotimer32_t;
#endif

typedef struct Timer {
	double time0, elapsed;
	char running;
	
	int fd;
#if defined (IRIX) || defined( IRIX64 )
	unsigned long long counter64;
	unsigned int counter32;
	unsigned int cycleval;
	
	volatile iotimer64_t *iotimer_addr64;
	volatile iotimer32_t *iotimer_addr32;

	void *unmapLocation;
	int unmapSize;
#elif defined(WINDOWS)
	LARGE_INTEGER performance_counter, performance_frequency;
	double one_over_frequency;
#elif defined( Linux )
	struct timeval timeofday;
#endif
} WireGLTimer;

UTIL_DECL WireGLTimer *wireGLTimerNewTimer( void );
UTIL_DECL void wireGLDestroyTimer( WireGLTimer *t );
UTIL_DECL void wireGLStartTimer( WireGLTimer *t );
UTIL_DECL void wireGLStopTimer( WireGLTimer *t );
UTIL_DECL void wireGLResetTimer( WireGLTimer *t );
UTIL_DECL double wireGLTimerTime( WireGLTimer *t );

typedef enum {
	TO_SCREEN,
	TO_FILE
} WireGLInstrumentationType;

//#define INSTRUMENT

UTIL_DECL void wireGLInitInstrumentation( WireGLInstrumentationType type, ...);
UTIL_DECL void wireGLInstrumentToScreen();
UTIL_DECL void wireGLInstrumentToFile( char *fname );
UTIL_DECL void wireGLInstrumentShutDown();
static inline void __wireGLInstrumentNOP( ... ) {}

#if defined(IGNORE_INSTRUMENT_FLAG) || defined(INSTRUMENT)

UTIL_DECL void wireGLInstrumentRegisterCounter( int id, char *name );
UTIL_DECL void wireGLInstrumentRegisterTimer( int id, char *name );
UTIL_DECL void wireGLInstrumentRegisterPerFrameCounter( int id, char *name );
UTIL_DECL void wireGLInstrumentRegisterPerServerCounter( int id, int num, char *name );

UTIL_DECL void wireGLInstrumentIncrPerFrameCounter( int id, float amount );
UTIL_DECL void wireGLInstrumentIncrPerServerCounter( int id, int which, float amount );
UTIL_DECL void wireGLInstrumentIncrCounter( int id, float amount );
UTIL_DECL void wireGLInstrumentSetCounter( int id, float amount );

UTIL_DECL void wireGLInstrumentResetTimer( int id );
UTIL_DECL void wireGLInstrumentStartTimer( int id );
UTIL_DECL void wireGLInstrumentStopTimer( int id );

UTIL_DECL float wireGLInstrumentGetCounter( int id );
UTIL_DECL double wireGLInstrumentGetTime( int id );

UTIL_DECL void wireGLInstrumentPrintPerServerCounter( int id );
UTIL_DECL void wireGLInstrumentPrintPerFrameCounter( int id );
UTIL_DECL void wireGLInstrumentPrintCounter( int id );
UTIL_DECL void wireGLInstrumentPrintTimer( int id );
UTIL_DECL void wireGLInstrumentNextFrame();

UTIL_DECL void wireGLInstrumentPrintf( const char *format, ... );

UTIL_DECL void wireGLInstrumentPrintAll( );

#else

#define wireGLInstrumentRegisterCounter( id, name ) do {} while(0)
#define wireGLInstrumentRegisterTimer( id, name )  do {} while(0)
#define wireGLInstrumentRegisterPerServerCounter( id, num, name )  do {} while(0)
#define wireGLInstrumentRegisterPerFrameCounter( id, name )  do {} while(0)
#define wireGLInstrumentIncrPerServerCounter( id, val, name )  do {} while(0)
#define wireGLInstrumentIncrPerFrameCounter( id, val )  do {} while(0)
#define wireGLInstrumentIncrCounter( id, amount )  do {} while(0)
#define wireGLInstrumentSetCounter( id, amount )  do {} while(0)

#define wireGLInstrumentResetTimer( id )  do {} while(0)
#define wireGLInstrumentStartTimer( id )  do {} while(0)
#define wireGLInstrumentStopTimer( id )  do {} while(0)

#define wireGLInstrumentGetCounter( id )  do {} while(0)
#define wireGLInstrumentGetTime( id )  0

#define wireGLInstrumentPrintCounter( id )  do {} while(0)
#define wireGLInstrumentPrintTimer( id )  do {} while(0)

#define wireGLInstrumentNextFrame() do {} while(0)

#define wireGLInstrumentPrintPerServerCounter( id )  do {} while(0)
#define wireGLInstrumentPrintPerFrameCounter( id )  do {} while(0)
#define wireGLInstrumentPrintf __wireGLInstrumentNOP

#define wireGLInstrumentPrintAll( )  do {} while(0)
#endif

UTIL_DECL int wireGLParseURL( char *url, char *protocol, char *hostname, int *port, int default_port );

UTIL_DECL char *wireGLGetLine( FILE *fp, char *buf );

#ifndef WINDOWS
#define WIREGL_PATH_SEPARATOR ":"
#else
#define WIREGL_PATH_SEPARATOR ";"
#endif

#define WIREGL_LITTLE_ENDIAN 0
#define WIREGL_BIG_ENDIAN 1

typedef struct {
	char dllname[1024];
	int refcnt;
#if defined(WINDOWS)
	HINSTANCE hinstLib;
#elif defined(IRIX) || defined(IRIX64) || defined( Linux )
	void *hinstLib;
#else
#error DSOs, please
#endif
} WireGLDSO;

UTIL_DECL WireGLDSO *wireGLDSOOpen( const char *dllname );
UTIL_DECL void *wireGLDSOGet( WireGLDSO *dso, const char *symname );
UTIL_DECL void *wireGLDSOGetNoError( WireGLDSO *dso, const char *symname );
UTIL_DECL void wireGLDSOClose( WireGLDSO *dso );

typedef struct {
		char *str;
		void (*parser)( char *buf, void *arg );
} WireGLParseStruct;

UTIL_DECL void wireGLParseFile( char *fname, WireGLParseStruct *ps, void *arg );

#endif /* WIREGL_UTIL_H */
