/*
** buffer.c
** http://graphics.stanford.edu/software/wiregl
**
** Copyright  2000 The Board of Trustees of The Leland Stanford Junior
** University. All rights reserved.
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or modify
** this software, provided that the above copyright notice and this
** permission notice appear in all copies of this software.  No part of
** this software or any derivatives thereof may be used in graphics systems
** for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind, 
** express, implied or otherwise, including without limitation, any warranty 
** of merchantability or fitness for a particular purpose.
**
*/

#include <stdio.h>

#include "wiregl/include/wiregl_protocol.h"
#include "wiregl/include/wiregl_util.h"
#include "wiregl/include/wiregl_client.h"
#include "wiregl/include/wiregl_instrument.h"
#include "geometry.h"

extern char *glNames[];

static enum { GEOMETRY, STATE, FORCE_ALL } flush_mode = GEOMETRY;
static WireGLPipeServer *state_server;

GLbitvalue current_bitmask;
WireGLOpcode force_opcode;

void wireGLGMDeath()
{
	WireGLOpcode opcode = WIREGL_GMDEATH_OPCODE;

	for (int i = 0 ; i < num_pipe_servers ; i++)
	{
		if (pipe_servers[i].connection->type == GM)
		{
			pipe_servers[i].connection->wireGLRegisterMemoryForDMA( 
					pipe_servers[i].connection, &opcode, 1 );
			pipe_servers[i].connection->wireGLInjectMessage( 
					pipe_servers[i].connection, &opcode, 1 );
			pipe_servers[i].connection->wireGLInjectMessage( 
					pipe_servers[i].connection, &opcode, 1 );
		}
	}
	wireGLSleep( 100 );
}

void wireGLFlushAll( GLcontext *virtual_context, WireGLOpcode op )
{
	flush_mode = FORCE_ALL;
	force_opcode = op;
	wireGLFlushBuffer( virtual_context );
	flush_mode = GEOMETRY;
#ifdef INSTRUMENT
	if (print_frame_rate)
	{
		double frame_time = wireGLInstrumentGetTime( WIREGL_FRAME_TIMER );
		wireGLInstrumentIncrPerFrameCounter( WIREGL_FRAME_TIMER, 1.0/frame_time );
		wireGLInstrumentResetTimer( WIREGL_FRAME_TIMER );
		wireGLInstrumentStartTimer( WIREGL_FRAME_TIMER );
	}
	wireGLInstrumentNextFrame();
#endif
}

void wireGLFlushBuffer( GLcontext *virtual_context )
{
	if (flush_mode != FORCE_ALL) // FORCE_ALL happens on SwapBuffers, so there may be nothing in any buffer.
	{
		if (pipe_server_opcode_current == pipe_server_opcode_start)
			return;
	}

	if (flush_mode == GEOMETRY || flush_mode == FORCE_ALL)
	{

#ifdef INSTRUMENT
		if (print_flush_time)
		{
			wireGLInstrumentStartTimer( WIREGL_FLUSH_TIMER );
		}
#endif
		// First, call the bucketer to find out where we're going.
		current_bitmask =__glbucket_flush();

		int i;
		int was_force = (flush_mode == FORCE_ALL);

		// Save away the pointers to the geometry buffer so we can copy it later

		unsigned char *geom_data_start = pipe_server_data_start;
		unsigned char *geom_data_current = pipe_server_data_current;
		unsigned char *geom_data_end = pipe_server_data_end;

		unsigned char *geom_opcode_start = pipe_server_opcode_start;
		unsigned char *geom_opcode_current = pipe_server_opcode_current;

		flush_mode = STATE;

		// Compress the geometry data

		//__glCompressGeometryData(pipe_server_data_start, pipe_server_data_current, pipe_server_data_end,
		//	pipe_server_opcode_start, pipe_server_opcode_current);

#ifdef INSTRUMENT
		if (print_min_bytes_sent)
		{
			wireGLInstrumentIncrCounter( WIREGL_MIN_BYTES_SENT, 
					(geom_data_current - geom_data_start) + 
					(geom_opcode_current - geom_opcode_start) );
		}
#endif
		for (i = 0 ; i < num_pipe_servers ; i++)
		{
			state_server = pipe_servers + i;

			// If the bucketer said we're sending to server "i", do it.

			if (current_bitmask & (1 << i))
			{

				// Cook the global pointers so they point at the buffers for this pipe server.
				pipe_server_data_start = state_server->data_start;
				pipe_server_data_end = state_server->data_end;
				pipe_server_data_current = state_server->data_current;

				pipe_server_opcode_start = state_server->opcode_start;
				pipe_server_opcode_current = state_server->opcode_current;

				// Generate the lazy state evaluation commands into the command buffer


#ifdef INSTRUMENT
				if (print_diff_context_time)
				{
					wireGLInstrumentStartTimer( WIREGL_DIFF_CONTEXT_TIMER );
				}

#endif
				__glDiffContext( state_server->context, virtual_context );

#ifdef INSTRUMENT
				if (print_diff_context_time)
				{
					wireGLInstrumentStopTimer( WIREGL_DIFF_CONTEXT_TIMER );
					wireGLInstrumentIncrPerFrameCounter( WIREGL_DIFF_CONTEXT_TIMER, wireGLInstrumentGetTime( WIREGL_DIFF_CONTEXT_TIMER ) );
					wireGLInstrumentResetTimer( WIREGL_DIFF_CONTEXT_TIMER );
				}

				if (print_min_bytes_sent)
				{
					wireGLInstrumentIncrCounter( WIREGL_MIN_BYTES_SENT, 
							(pipe_server_data_current - state_server->data_current) + 
							(pipe_server_opcode_current - state_server->opcode_current));
				}
#endif
				// If there's room to slap the geometry onto the end of the server's buffer, do it.
				if (pipe_server_data_end - pipe_server_data_current > geom_data_current - geom_data_start )
				{
					memcpy( pipe_server_data_current, geom_data_start, (geom_data_current - geom_data_start) );
					state_server->data_current = pipe_server_data_current + (geom_data_current - geom_data_start);

					memcpy( pipe_server_opcode_current, geom_opcode_start, (geom_opcode_current - geom_opcode_start) );
					state_server->opcode_current = pipe_server_opcode_current + (geom_opcode_current - geom_opcode_start);
				}
				else
				{
					// Otherwise, time to get it out the door early.
				
					wireGLFlushBuffer( virtual_context );
					pipe_server_data_start = geom_data_start;
					pipe_server_data_current = geom_data_current;
					pipe_server_opcode_start = geom_opcode_start;
					pipe_server_opcode_current = geom_opcode_current;
					wireGLFlushBuffer( virtual_context );
					state_server->data_current = state_server->data_start;
					state_server->opcode_current = state_server->opcode_start;
				}
			}
		}

		if (was_force)
		{
			// If someone called Swap Buffers, we need to get the swapbuffers opcode onto the end
			// of each state buffer, and flush it.
			
			for (i = 0 ; i < num_pipe_servers ; i++)
			{
				state_server = pipe_servers + i;

				pipe_server_data_start = state_server->data_start;
				pipe_server_data_end = state_server->data_end;
				pipe_server_data_current = state_server->data_current;

				pipe_server_opcode_start = state_server->opcode_start;
				pipe_server_opcode_current = state_server->opcode_current;

				__glCloseBeginEnd(state_server->context);
				WRITE_OPCODE( force_opcode );

				wireGLFlushBuffer( virtual_context );
				state_server->data_current = state_server->data_start;
				state_server->opcode_current = state_server->opcode_start;
			}
		}

		// We're done -- reset all the global pointers so that subsequent geometry commands
		// will pack into the right place.

		pipe_server_data_start = geom_data_start;
		pipe_server_data_current = geom_data_start;
		pipe_server_data_end = geom_data_end;
		pipe_server_opcode_start = geom_opcode_start;
		pipe_server_opcode_current = geom_opcode_start;
		flush_mode = GEOMETRY;
		current_bitmask = ~0;

		// Let the bucketer know that we're done with this block
		__glbucket_endflush();

#ifdef INSTRUMENT
		if (print_flush_time)
		{
			wireGLInstrumentStopTimer( WIREGL_FLUSH_TIMER );
					wireGLInstrumentIncrPerFrameCounter( WIREGL_FLUSH_TIMER, wireGLInstrumentGetTime( WIREGL_FLUSH_TIMER ) );
			wireGLInstrumentResetTimer( WIREGL_FLUSH_TIMER );
		}
#endif

	}
	else
	{
		if (pipe_server_opcode_current == pipe_server_opcode_start)
			return;

		// The actual transmission code -- flush *one* server buffer.  The global
		// variable "state_server" is which one to send to.

		int opcode_len = pipe_server_opcode_current - pipe_server_opcode_start;
		int data_len = pipe_server_data_current - pipe_server_data_start;

		if (state_server->connection->use_async)
		{
			wireGLSimpleError( "Asyncronous connections don't work with bucketing..." );
		}
		else
		{

			// Send everything in two chunks -- control followed by data
			
#ifdef INSTRUMENT
			if (print_send_time)
			{
				wireGLInstrumentStartTimer( WIREGL_SEND_TIMER );
			}
#endif
			state_server->connection->wireGLSendMessage( state_server->connection, pipe_server_opcode_start, opcode_len );
			state_server->connection->wireGLSendMessage( state_server->connection, pipe_server_data_start, data_len );

#ifdef INSTRUMENT
			if (print_send_time)
			{
				wireGLInstrumentStopTimer( WIREGL_SEND_TIMER );
					wireGLInstrumentIncrPerFrameCounter( WIREGL_SEND_TIMER, wireGLInstrumentGetTime( WIREGL_SEND_TIMER ) );
				wireGLInstrumentResetTimer( WIREGL_SEND_TIMER );
			}
			if (print_total_bytes_sent)
			{
				wireGLInstrumentIncrCounter( WIREGL_TOTAL_BYTES_SENT, opcode_len + data_len );
			}
			if (print_bytes_per_server)
			{
				wireGLInstrumentIncrPerServerCounter( 
						WIREGL_BYTES_PER_SERVER, state_server->index, 
						opcode_len + data_len );
			}
			if (print_bytes_per_frame_per_server)
			{
				wireGLInstrumentIncrPerFrameCounter( 
						WIREGL_BYTES_PER_SERVER_PER_FRAME + state_server->index, 
						opcode_len + data_len );
			}
#endif
			state_server->data_current = state_server->data_start;
			state_server->opcode_current = state_server->opcode_start;

			// need to do this in case we get a "too big" packet.
			pipe_server_data_current = state_server->data_start;
			pipe_server_opcode_current = state_server->opcode_start;

		}

#if 0
		{
			int i;
			for (i = 0 ; i < opcode_len ; i++)
			{
				printf ("%d/%d: %s\n", i+1, opcode_len, glNames[pipe_server_opcode_start[i]] );
			}
		}
#endif

	}
}

void *wireGLAllocPacket( unsigned int size )
{
	unsigned char *data_ptr;
	// Maybe it will fit...
	int data_len = pipe_server_data_current - pipe_server_data_start;
	if (data_len + size < pipe_server_buffer_size)
	{
		GET_BUFFERED_POINTER( size );
		return data_ptr;
	}

	// Okay, it didn't fit.  Maybe it will after we flush.
	wireGLFlushBuffer( pipe_server_state );
	data_len = pipe_server_data_current - pipe_server_data_start;
	if (data_len + size < pipe_server_buffer_size)
	{
		GET_BUFFERED_POINTER( size );
		return data_ptr;
	}

	// It's really way too big.
	return wireGLAlloc( size );
}
#define IS_BUFFERED( packet ) \
	((packet) >= pipe_server_data_start && \
	 (packet) < pipe_server_data_start + pipe_server_buffer_size)

void wireGLSendPacket( unsigned char opcode, void *packet )
{
	if (IS_BUFFERED( packet ))
	{
		WRITE_OPCODE( opcode );
		return;
	}
	int *len = (int *) packet;
	int i;

	static char first_packet;
	static int dma_initialized = 0;

	if (!dma_initialized)
	{
		for (i = 0 ; i < num_pipe_servers; i++)
		{
			pipe_servers[i].connection->wireGLRegisterMemoryForDMA( pipe_servers[i].connection, &first_packet, 1 );
		}
		dma_initialized = 1;
	}

	first_packet = opcode;

	state_server->connection->wireGLRegisterMemoryForDMA( state_server->connection, packet, *len );
	if (state_server->connection->use_async)
	{
		wireGLSimpleError( "Asyncronous connections don't work with bucketing..." );
	}
	else
	{
		if (print_send_time)
		{
			wireGLInstrumentStartTimer( WIREGL_SEND_TIMER );
		}

		state_server->connection->wireGLSendMessage( 
				state_server->connection, &first_packet, 1 );
		state_server->connection->wireGLSendMessage( 
			state_server->connection, packet, *len ); 
		state_server->connection->wireGLDeregisterMemoryForDMA( 
			state_server->connection, packet, *len );

		if (print_send_time)
		{
			wireGLInstrumentStopTimer( WIREGL_SEND_TIMER );
					wireGLInstrumentIncrPerFrameCounter( WIREGL_SEND_TIMER, wireGLInstrumentGetTime( WIREGL_SEND_TIMER ) );
			wireGLInstrumentResetTimer( WIREGL_SEND_TIMER );
		}
		if (print_total_bytes_sent)
		{
			wireGLInstrumentIncrCounter( WIREGL_TOTAL_BYTES_SENT, 1 + *len );
		}

		if (print_bytes_per_server)
		{
			wireGLInstrumentIncrPerServerCounter( 
					WIREGL_BYTES_PER_SERVER, state_server->index, 
					1 + *len );
		}
		if (print_bytes_per_frame_per_server)
		{
			wireGLInstrumentIncrPerFrameCounter( 
					WIREGL_BYTES_PER_SERVER_PER_FRAME + state_server->index, 
					1 + *len );
		}
		if (print_min_bytes_sent)
		{
			wireGLInstrumentIncrCounter( WIREGL_MIN_BYTES_SENT, 1 + *len );
		}
	}
#if 0
	printf ("big: %s\n", glNames[first_packet] );
#endif
}

void wireGLFreePacket( void *packet )
{
	if (IS_BUFFERED( packet ) )
		return;
	wireGLFree( packet );
}
