#include "wiregl/include/wireGL.h"
#include "wiregl/include/wiregl_protocol.h"
#include "wiregl/include/wiregl_pipe.h"
#include "wiregl/include/wiregl_common.h"

#if NORMAL_COMPRESSION

/* Structures
 * ==========
 *
 * CompressedNormalT
 * -----------------
 * This structure stores the compressed data that is produced by the compression
 * algorithm. The symmetry field contains information regarding in which octant and
 * sextant the normal lies, and the left and right values contain a compressed
 * version of nx/nz and ny/nz.
 *
 * NormalDataT
 * -----------
 * This type is used to force the compiler to write the CompressedNormalT to the
 * data buffer as a specific number of bytes. Normally, the CompressedNormalT would
 * take up four bytes even if only the first two are being used. So, we declare a
 * separate type that takes up the amount of room we want to use in the data buffer
 * and use a pointer cast to write only that much data.
 *
 * COMPRESSION_SHIFT
 * -----------------
 * This value dictates how many bits should be removed from the mantissa of the
 * floating point value in the quantization step.
 */

#if TWO_BYTE_COMPRESSION

typedef struct
{
	unsigned symmetry : 6;
	unsigned leftValue : 5;
	unsigned rightValue : 5;
} CompressedNormalT;

typedef short NormalDataT;

#define COMPRESSION_SHIFT 18

#else

typedef struct
{
	unsigned char symmetry;
	unsigned char leftValue;
	unsigned char rightValue;
} CompressedNormalT;

typedef long NormalDataT;

#define COMPRESSION_SHIFT 15

#endif

#include <math.h>
#include "floatlib.c"

#define NORMAL_SIZE sizeof(NormalDataT)
#define NEGATIVE_X 0x0001
#define NEGATIVE_Y 0x0002
#define NEGATIVE_Z 0x0004
#define SEXTANT_XY 0x0008
#define SEXTANT_XZ 0x0010
#define SEXTANT_YZ 0x0020

/* swap
 * ----
 * This function swaps two floating point values.
 */

static __inline void swap(GLfloat *in1, GLfloat *in2)
{
	GLfloat temp;
	temp = *in1;
	*in1 = *in2;
	*in2 = temp;
}

/* __retrieveValues
 * ----------------
 * This function decompresses the values sent by the WireGL client. The data still needs to be
 * sorted and have its signs restored after this function.
 */

static __inline void __retrieveValues ( unsigned char leftMantissa, unsigned char rightMantissa, GLfloat * outValues )
{
	GLfloat nxnz, nynz;
	unsigned long longMantissa;

	longMantissa = leftMantissa << COMPRESSION_SHIFT;
	nxnz = __normalizeMantissa(longMantissa);

	longMantissa = rightMantissa << COMPRESSION_SHIFT;
	nynz = __normalizeMantissa(longMantissa);

	/*           sqrt(nx * nx + ny * ny + nz * nz) = 1
	 *               (nx * nx + ny * ny + nz * nz) = 1
	 *         ((nx / nz) ^ 2 + (ny / nz) ^ 2 + 1) = 1 / nz ^ 2
	 * 1 / sqrt((nx / nz) ^ 2 + (ny / nz) ^ 2 + 1) = nz
	 */
	
	outValues[2] = 1 / sqrtf(nxnz * nxnz + nynz * nynz + 1);
	outValues[1] = nynz * outValues[2];
	outValues[0] = nxnz * outValues[2];
}

/* __assignOctantAndSextant
 * ------------------------
 * This function restores the order and signs of the values given the symmetry data passed to
 * it by the WireGL client.
 */

static __inline void __assignOctantAndSextant ( unsigned char symmetry, GLfloat * outValues )
{
	if(symmetry & SEXTANT_XY)							// x < y
	{
		if(symmetry & SEXTANT_XZ)						// x < y && x < z
		{
			if(!(symmetry & SEXTANT_YZ))				// x < z < y
				swap( &outValues[1], &outValues[2] );
		}
		else											// z < x < y
			swap( &outValues[1], &outValues[2] );
	}
	else												// y < x
	{
		if(symmetry & SEXTANT_YZ)						// y < x && y < z
		{
			if(!(symmetry & SEXTANT_XZ))				// y < z < x
				swap( &outValues[0], &outValues[2] );
		}
		else											// z < y < x
			swap( &outValues[0], &outValues[2] );
	}

	if(symmetry & NEGATIVE_X)
		outValues[0] = -outValues[0];
	if(symmetry & NEGATIVE_Y)
		outValues[1] = -outValues[1];
	if(symmetry & NEGATIVE_Z)
		outValues[2] = -outValues[2];
}

void __decodeglNormal3f( void )
{
	GLfloat v[3];
	NormalDataT inData;
	CompressedNormalT *inNormal = (CompressedNormalT*)&inData;

	inData = READ_DATA(0, NormalDataT);

	/* Retrieve the values, restore the symmetry and call glNormal. */
	
	__retrieveValues(inNormal->leftValue, inNormal->rightValue, v);
	__assignOctantAndSextant(inNormal->symmetry, v);
	glNormal3fv( ( GLfloat* ) v );

	INCR_DATA_PTR( NORMAL_SIZE );
}

void __decodeglNormal3d( void )
{
	glNormal3dv( ( GLdouble* ) unpacker_data_ptr );
	INCR_DATA_PTR( 24 );
}
void __decodeglNormal3i( void )
{
	glNormal3iv( ( GLint* ) unpacker_data_ptr );
	INCR_DATA_PTR( 12 );
}

#else

void __decodeglNormal3d( void )
{
	glNormal3dv( ( GLdouble* ) unpacker_data_ptr );
	INCR_DATA_PTR( 24 );
}
void __decodeglNormal3f( void )
{
	glNormal3fv( ( GLfloat* ) unpacker_data_ptr );
	INCR_DATA_PTR( 12 );
}
void __decodeglNormal3i( void )
{
	glNormal3iv( ( GLint* ) unpacker_data_ptr );
	INCR_DATA_PTR( 12 );
}

#endif