/*
** error.c
** http://graphics.stanford.edu/software/wiregl
**
** Copyright  2000 The Board of Trustees of The Leland Stanford Junior
** University. All rights reserved.
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or modify
** this software, provided that the above copyright notice and this
** permission notice appear in all copies of this software.  No part of
** this software or any derivatives thereof may be used in graphics systems
** for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind, 
** express, implied or otherwise, including without limitation, any warranty 
** of merchantability or fitness for a particular purpose.
**
*/

#include "__utilsource.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#if !defined(WINDOWS)
#include <unistd.h>
#else
#include <assert.h>
#include <windows.h>
#include <process.h>
#endif
#include "wiregl/include/wiregl_util.h"

WireGLSpinLock output_sema;

void wireGLIntError( const char *format, int core, int print, va_list args ) NORETURN;

void wireGLIntError( const char *format, int core, int print, va_list args ) 
{
	static char txt[8092];
	int offset;
	wireGLAcquireSpinLock( &output_sema );
	offset = sprintf( txt, "WireGL Error (%d): ", (int)getpid());
	vsprintf( &(txt[offset]), format, args );
#ifdef WINDOWS
	if (!print)
		MessageBox( NULL, txt, "wireGL error", MB_OK );
	else
		fprintf( stderr, "%s\n", txt );
#else
	fprintf( stderr, "%s\n", txt );
#endif

	wireGLReleaseSpinLock( &output_sema );
	if (core)
		abort();
	else
		exit(0);

}

UTIL_DECL void wireGLError( const char *format, ... )
{
		static char *do_print = getenv( "WIREGL_GUI_ERROR" );
	va_list args;
	va_start( args, format );
	wireGLIntError( format, 1, (do_print == NULL), args );
	va_end( args );
}

UTIL_DECL void wireGLSimpleError( const char *format, ... )
{
	static char *do_print = getenv( "WIREGL_GUI_ERROR" );
	va_list args;
	va_start( args, format );
	wireGLIntError( format, 0, (do_print == NULL), args );
	va_end( args );
}

UTIL_DECL void wireGLPrintError( const char *format, ... )
{
	va_list args;
	va_start( args, format );
	wireGLIntError( format, 0, 1, args );
	va_end( args );
}

#ifndef WIREGL_NO_WARN

static WireGLWarningLevel warn_level = WIREGL_WARN_DEBUG;

UTIL_DECL void wireGLWarning( WireGLWarningLevel level, const char *format, ... )
{
	if (level <= warn_level)
	{
		va_list args;
		va_start( args, format );
		wireGLAcquireSpinLock( &output_sema );
		fprintf ( stderr, "WireGL Warning (%d): ", (int)getpid());
		vfprintf( stderr, format, args );
		fprintf( stderr, "\n");
		wireGLReleaseSpinLock( &output_sema );
		va_end( args );
	}
}

#endif

UTIL_DECL void wireGLInitWarnings( void )
{
  char *warn_str;

  warn_level = WIREGL_WARN_CRITICAL;
  warn_str = getenv( "WARN_LEVEL" );
  if (warn_str)
    {
      if (!strcmp( warn_str, "VERBOSE_DEBUG" ) )
        warn_level = WIREGL_WARN_VERBOSE_DEBUG;
      else if (!strcmp( warn_str, "DEBUG" ) )
        warn_level = WIREGL_WARN_DEBUG;
      else if (!strcmp( warn_str, "CHATTY" ) )
        warn_level = WIREGL_WARN_CHATTY;
      else if (!strcmp( warn_str, "NOTICE" ) )
        warn_level = WIREGL_WARN_NOTICE;
      else if (!strcmp( warn_str, "IMPORTANT" ) )
        warn_level = WIREGL_WARN_IMPORTANT;
      else warn_level = WIREGL_WARN_CRITICAL;
    }
  wireGLInitSpinLock( &output_sema );
}
