/*
** gm.c
** http://graphics.stanford.edu/software/wiregl
**
** Copyright  2000 The Board of Trustees of The Leland Stanford Junior
** University. All rights reserved.
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or modify
** this software, provided that the above copyright notice and this
** permission notice appear in all copies of this software.  No part of
** this software or any derivatives thereof may be used in graphics systems
** for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind, 
** express, implied or otherwise, including without limitation, any warranty 
** of merchantability or fitness for a particular purpose.
**
*/

#include "__utilsource.h"

#include <stdio.h>
#include <stdlib.h>
#include <gm.h>

#include "wiregl/include/wiregl_util.h"
#include "wiregl/include/wiregl_common.h"

#define GM_MSG_SIZE 24

static void wiregl_gm_wait_for_it( WireGLConnection *conn )
{
	while (conn->wiregl_gm_spinning_for_fun)
	{
		gm_recv_event *event = gm_receive( conn->wiregl_gm_port );
		gm_unknown( conn->wiregl_gm_port, event );
	}
	conn->wiregl_gm_spinning_for_fun = 1;
}

static void wiregl_gm_finish_callback( struct gm_port *p, void *context, gm_status_t status )
{
	WireGLConnection *conn = (WireGLConnection *) context;
	if (status != GM_SUCCESS)
	{
		wireGLSimpleError( "Bad GM Status code in the callback: %d", status );
	}
	conn->wiregl_gm_spinning_for_fun = 0;
}

#ifndef min
#define min(a,b) ((a) < (b) ? (a) : (b))
#endif

// We don't really read "exact" on GM -- we rely on the fact that we know what our
// packets look like.  This is awful.

UTIL_DECL int wireGLGMReadMessage( WireGLConnection *conn, void *buf )
{
	gm_recv_event *event;
	
	gm_provide_receive_buffer( conn->wiregl_gm_port, buf, GM_MSG_SIZE, GM_HIGH_PRIORITY );

	for (;;)
	{
		event = gm_blocking_receive_no_spin ( conn->wiregl_gm_port );
		switch( event->recv.type )
		{
			case GM_HIGH_RECV_EVENT:
				return ntohl( event->recv.length );
			default:
				gm_unknown( conn->wiregl_gm_port, event );
				break;
		}
	}
	return 0;
}

UTIL_DECL void wireGLGMInjectMessage( WireGLConnection *conn, void *buf, int bytes )
{
	if (bytes == 0)
	{
		// we can't have a zero length message in GM.  However, the only time the system
		// will try to send a zero length message is when we have packets that are all opcodes and no
		// data.  In this case, the data will be ignored by the server, but we have to send SOMETHING,
		// so se send one byte of crap.
		gm_send_with_callback( conn->wiregl_gm_port, conn->wiregl_gm_zero_length_hack, GM_MSG_SIZE, 1, 
				GM_HIGH_PRIORITY, conn->wiregl_gm_server_id, WIREGL_GM_PORT, wiregl_gm_finish_callback, conn );
	}
	else
	{
		gm_send_with_callback( conn->wiregl_gm_port, buf, GM_MSG_SIZE, bytes, GM_HIGH_PRIORITY, 
				conn->wiregl_gm_server_id, WIREGL_GM_PORT, wiregl_gm_finish_callback, conn );
	}
}

UTIL_DECL void wireGLGMSendMessage( WireGLConnection *conn, void *buf, int bytes )
{
	wireGLGMInjectMessage( conn, buf, bytes );
	wiregl_gm_wait_for_it( conn );
}

UTIL_DECL void wireGLGMRegisterMemoryForDMA( WireGLConnection *conn, void *buf, int len )
{
	gm_register_memory( conn->wiregl_gm_port, buf, len );
}

UTIL_DECL void wireGLGMDeregisterMemoryForDMA( WireGLConnection *conn, void *buf, int len )
{
	gm_deregister_memory( conn->wiregl_gm_port, buf, len );
}

UTIL_DECL void wireGLStartGM( WireGLConnection *conn )
{
	static int gm_initialized = 0;

	conn->type = GM;
	wireGLCloseSocket( conn->tcp_socket );

	conn->wireGLRegisterMemoryForDMA = wireGLGMRegisterMemoryForDMA;
	conn->wireGLDeregisterMemoryForDMA = wireGLGMDeregisterMemoryForDMA;
	conn->wireGLReadMessage = wireGLGMReadMessage;
	conn->wireGLSendMessage = wireGLGMSendMessage;
	conn->wireGLInjectMessage = wireGLGMInjectMessage;

	if (!gm_initialized)
	{
		gm_init();
		gm_initialized = 1;
	}

	if (gm_open( &(conn->wiregl_gm_port), 0, WIREGL_GM_PORT, "WireGL GM Server", GM_API_VERSION_1_0 ) != GM_SUCCESS)
	{
		wireGLSimpleError( "Can't open the GM port on %d", WIREGL_GM_PORT );
	}

	conn->wiregl_gm_spinning_for_fun = 1;
}

UTIL_DECL void wireGLMakeGM( WireGLConnection *conn )
{
	static int gm_initialized = 0;
	static struct gm_port *the_gm_port;

	conn->type = GM;
	wireGLCloseSocket( conn->tcp_socket );

	conn->wireGLRegisterMemoryForDMA = wireGLGMRegisterMemoryForDMA;
	conn->wireGLDeregisterMemoryForDMA = wireGLGMDeregisterMemoryForDMA;
	conn->wireGLReadMessage = wireGLGMReadMessage;
	conn->wireGLSendMessage = wireGLGMSendMessage;
	conn->wireGLInjectMessage = wireGLGMInjectMessage;

	if (!gm_initialized)
	{
		gm_init();
		gm_initialized = 1;
		if (gm_open( &(the_gm_port), 0, WIREGL_GM_PORT, "WireGL GM Client", GM_API_VERSION_1_0 ) != GM_SUCCESS)
		{
			wireGLSimpleError( "Can't open the GM port on %d", WIREGL_GM_PORT );
		}
	}
	conn->wiregl_gm_port = the_gm_port;

	if ( (conn->wiregl_gm_server_id = gm_host_name_to_node_id( conn->wiregl_gm_port, conn->hostname )) 
			== GM_NO_SUCH_NODE_ID )
	{
		wireGLSimpleError( "There's no host called %s on the Myrinet network", conn->hostname );
	}

	conn->wiregl_gm_spinning_for_fun = 1;
	conn->wiregl_gm_zero_length_hack = (char *) gm_dma_malloc( conn->wiregl_gm_port, 1 );
}

