/*
** mutex.c
** http://graphics.stanford.edu/software/wiregl
**
** Copyright  2000 The Board of Trustees of The Leland Stanford Junior
** University. All rights reserved.
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or modify
** this software, provided that the above copyright notice and this
** permission notice appear in all copies of this software.  No part of
** this software or any derivatives thereof may be used in graphics systems
** for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind, 
** express, implied or otherwise, including without limitation, any warranty 
** of merchantability or fitness for a particular purpose.
**
*/

#include "__utilsource.h"

#include "wiregl/include/wiregl_util.h"

UTIL_DECL void wireGLInitSemaphore( WireGLSemaphore *s, int value, char *name )
{
#if defined(IRIX) || defined( IRIX64 )
	s->sema = usnewsema( shared_mem, value );
	if (!s->sema) {
		wireGLError( "Couldn't allocate a semaphore" );
	}
#endif
#if defined(WINDOWS)
	if (value == 0) 
		s->sema = CreateSemaphore(NULL, value, 1000000, name);
	else 
		s->sema = CreateSemaphore(NULL, value, value, name);

	if (!s->sema)
		wireGLError( "Couldn't allocate a semaphore" );
#endif

#if defined(Linux)
	if(sem_init(&(s->sema), 0, value)) {
		wireGLError( "Couldn't allocate a semaphore");
	}
#endif
}

UTIL_DECL void wireGLAcquireSemaphore( WireGLSemaphore *s )
{
#if defined( IRIX ) || defined( IRIX64 )
	if (uspsema( s->sema ) == -1)
		wireGLError( "Error acquiring a semaphore" );
#endif
#if defined (WINDOWS)
	DWORD status = WaitForSingleObject(s->sema,INFINITE);

	/* Do I need to check for any other status's for failure? */
	if (status == WAIT_FAILED) {
		wireGLError("Error acquiring a semaphore");
	}
#endif
#if defined(Linux)
	if((sem_wait(&(s->sema))))
		wireGLError( "Error acquiring a semaphore");

#endif
}

UTIL_DECL void wireGLReleaseSemaphore( WireGLSemaphore *s )
{
#if defined( IRIX ) || defined( IRIX64 )
	if (usvsema( s->sema ) == -1)
		wireGLError( "Error releasing a semaphore" );
#endif

#if defined (WINDOWS)
	if (!ReleaseSemaphore(s->sema, 1, NULL)) 
		wireGLError("Error releasing a semaphore:%d",GetLastError());
#endif

#if defined (Linux)
	if(sem_post(&(s->sema)))
		wireGLError("Error releasing a semaphore");
#endif
}

UTIL_DECL void wireGLDestroySemaphore( WireGLSemaphore *s )
{
#if defined( IRIX ) || defined( IRIX64 )
	usfreesema( s->sema, shared_mem );
#endif

#if defined (WINDOWS)
	CloseHandle(s->sema);
#endif
#if defined (Linux)
	sem_destroy(&(s->sema));
#endif
}

UTIL_DECL void wireGLInitSpinLock( WireGLSpinLock *sl )
{
	*sl = 0;
}

UTIL_DECL void wireGLAcquireSpinLock( WireGLSpinLock *sl )
{
#ifdef WINDOWS
	LONG result = InterlockedExchange( (long *) sl, 1 );
	while (result)
	{
		while (*sl)
			;
		result = InterlockedExchange( (long *) sl, 1 );
	}
#endif
}

UTIL_DECL void wireGLReleaseSpinLock( WireGLSpinLock *sl )
{
	*sl = 0;
}

UTIL_DECL void wireGLInitCondVar( WireGLCondVar *cv, WireGLSemaphore *threadSema, volatile int *flag )
{
	cv->threadSema = threadSema;
	cv->someone_waiting = flag;
} 

UTIL_DECL void wireGLWaitCondVar( WireGLCondVar *cv, WireGLSpinLock *sl )
{
	*(cv->someone_waiting) = 1;
	wireGLReleaseSpinLock( sl );
	wireGLAcquireSemaphore( cv->threadSema );
	wireGLAcquireSpinLock( sl );
} 

UTIL_DECL void wireGLSignalCondVar( WireGLCondVar *cv )
{
	if (*(cv->someone_waiting))
	{
		*(cv->someone_waiting) = 0;
		wireGLReleaseSemaphore( cv->threadSema );
	}
} 

