/*
** net.c
** http://graphics.stanford.edu/software/wiregl
**
** Copyright  2000 The Board of Trustees of The Leland Stanford Junior
** University. All rights reserved.
** Except for commercial resale, lease, license or other commercial
** transactions, permission is hereby given to use, copy, and/or modify
** this software, provided that the above copyright notice and this
** permission notice appear in all copies of this software.  No part of
** this software or any derivatives thereof may be used in graphics systems
** for resale or for use in a commercial product.
**
** This software is provided "as is" and without warranty of any kind, 
** express, implied or otherwise, including without limitation, any warranty 
** of merchantability or fitness for a particular purpose.
**
*/

#include "__utilsource.h"

#ifdef WINDOWS
#include <winsock.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "wiregl/include/wiregl_util.h"
#include "wiregl/include/wiregl_common.h"

int wireGLTCPIPReadMessage( WireGLConnection *, void *) ;
int wireGLFileReadMessage( WireGLConnection *, void *) ;
int wireGLDevNullReadMessage( WireGLConnection *, void *) ;

void wireGLTCPIPSendMessage( WireGLConnection *, void *, int ) ;
void wireGLFileSendMessage( WireGLConnection *, void *, int );
void wireGLDevNullSendMessage( WireGLConnection *, void *, int ) ;

void wireGLNopRegisterMemoryForDMA( WireGLConnection *, void *, int ) {}
void wireGLNopDeregisterMemoryForDMA( WireGLConnection *, void *, int ) {}

void wireGLErrorInjectMessage( WireGLConnection *, void *, int )
{
	wireGLSimpleError( "InjectMessage called on a connection that doesn't support it!" );
}

static void wireGLAsyncSender( void *arg )
{
	WireGLConnection *conn = (WireGLConnection *) arg;
	for (;;)
	{
		wireGLAcquireSemaphore( &(conn->data_ready) );
		conn->wireGLSendMessage( conn, conn->data_to_send, conn->data_length );
		if (conn->data_to_send2)
		{
			conn->wireGLSendMessage( conn, conn->data_to_send2, conn->data_length2 );
		}
		wireGLReleaseSemaphore( &(conn->data_sent ) );
	}
}

UTIL_DECL void wireGLSendAsync( WireGLConnection *conn, void *data, int length )
{
	wireGLAcquireSemaphore( &(conn->data_sent ) );
	conn->data_to_send = data;
	conn->data_length = length;
	conn->data_to_send2 = NULL;
	wireGLReleaseSemaphore( &(conn->data_ready ) );
}

UTIL_DECL void wireGLSendTwoAsync( WireGLConnection *conn, void *data, int length, void *data2, int length2 )
{
	wireGLAcquireSemaphore( &(conn->data_sent ) );
	conn->data_to_send = data;
	conn->data_length = length;
	conn->data_to_send2 = data2;
	conn->data_length2 = length2;
	wireGLReleaseSemaphore( &(conn->data_ready ) );
}

UTIL_DECL WireGLConnection *wireGLConnectToServer( char *server, int default_port, WireGLClientConnection *connection )
{
	struct sockaddr_in servaddr;
	struct hostent *hp;
	char hostname[4096], protocol[4096];
	char *async = getenv( "WIREGL_ASYNC_SEND" );
	int port;
	int window_size = 64*1024;

	WireGLClientConnection local_connection;
	if (!connection) 
		connection = &local_connection;

	WireGLConnection *conn = (WireGLConnection *) wireGLAlloc( sizeof( *conn ) );

	if (async)
	{
		wireGLInitSemaphore( &(conn->data_sent), 1, NULL );
		wireGLInitSemaphore( &(conn->data_ready), 0, NULL );
		wireGLCreateThread( wireGLAsyncSender, conn );
		conn->use_async = 1;
	}
	else
	{
		conn->use_async = 0;
	}

	if (!wireGLParseURL( server, protocol, hostname, &port, default_port ))
	{
		wireGLSimpleError( "Malformed URL: \"%s\"", server );
	}
	conn->hostname = strdup( hostname );

	conn->wireGLRegisterMemoryForDMA = wireGLNopRegisterMemoryForDMA;
	conn->wireGLDeregisterMemoryForDMA = wireGLNopDeregisterMemoryForDMA;
	conn->wireGLInjectMessage = wireGLErrorInjectMessage;

	connection->please_share_memory = 0;
	connection->please_use_gm = 0;
	conn->please_share_memory = 0;
	conn->please_use_gm = 0;

	if (!strcmp( protocol, "devnull" ) )
	{
		conn->type = DROP_PACKETS;
		conn->wireGLReadMessage = wireGLDevNullReadMessage;
		conn->wireGLSendMessage = wireGLDevNullSendMessage;
		return conn;
	}
	else if (!strcmp( protocol, "file" ) )
	{
		conn->type = FILE_TRACE;
		conn->fp = fopen( hostname, "wb" );
		if (conn->fp == NULL)
			wireGLSimpleError( "Couldn't open \"%s\"", hostname );
		conn->wireGLReadMessage = wireGLFileReadMessage;
		conn->wireGLSendMessage = wireGLFileSendMessage;
		return conn;
	}
	if (!strcmp( protocol, "gm" ))
	{
		connection->please_use_gm = 1;
		conn->please_use_gm = 1;
	}
	else if (strcmp( protocol, "tcpip" ))
	{
		wireGLSimpleError( "Unknown Protocol: \"%s\"", protocol );
	}

	conn->type = TCPIP;
	conn->tcp_socket = socket( AF_INET, SOCK_STREAM, 0 );
	if (conn->tcp_socket < 0)
		wireGLSimpleError( "socket error" );

	setsockopt( conn->tcp_socket, SOL_SOCKET, SO_SNDBUF, (char *) &window_size, sizeof( window_size ) );
	setsockopt( conn->tcp_socket, SOL_SOCKET, SO_RCVBUF, (char *) &window_size, sizeof( window_size ) );

	memset( &servaddr, 0, sizeof( servaddr ) );
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons( port );

	hp = gethostbyname( hostname );
	if (!hp)
	{
		wireGLSimpleError( "Unknown host: \"%s\"", hostname );
	}
	memcpy( (char *) &servaddr.sin_addr, hp->h_addr, sizeof( servaddr.sin_addr ) );
	if (connect( conn->tcp_socket, (struct sockaddr *) &servaddr,
				sizeof( servaddr )) < 0)
		wireGLSimpleError( "Couldn't connect to %s:%d using protocol \"%s\"", hostname, port, protocol );

	conn->wireGLReadMessage = wireGLTCPIPReadMessage;
	conn->wireGLSendMessage = wireGLTCPIPSendMessage;

	conn->wireGLSendMessage( conn, connection, sizeof( *connection ) );

	return conn;
}

UTIL_DECL void wireGLFinishConnectToServer( WireGLConnection *conn )
{
#ifdef GM_SUPPORT
	if (conn->please_use_gm)
		wireGLMakeGM( conn );
#endif
}

#ifdef WINDOWS
typedef int socklen_t;
#endif

UTIL_DECL WireGLServer *wireGLBecomeServer( int port )
{
	WireGLServer *server = (WireGLServer *) wireGLAlloc( sizeof( *server ) );

	int window_size = 64*1024;

	// we're not going to be doing anthing asyncronous on the server side.

	server->sock = socket( AF_INET, SOCK_STREAM, 0 );
	if (server->sock == -1)
	{
		wireGLSimpleError( "Couldn't create socket" );
	}
	setsockopt( server->sock, SOL_SOCKET, SO_SNDBUF, (char *) &window_size, sizeof( window_size ) );
	setsockopt( server->sock, SOL_SOCKET, SO_RCVBUF, (char *) &window_size, sizeof( window_size ) );


	server->addr.sin_family = AF_INET;
	server->addr.sin_addr.s_addr = INADDR_ANY;
	server->addr.sin_port = htons( port );
	if ( bind( server->sock, (struct sockaddr *) &server->addr, sizeof( server->addr ) ) == -1 )
	{
		wireGLSimpleError( "Couldn't bind to socket" );
	}
	return server;
}

UTIL_DECL WireGLConnection *wireGLAcceptClient( WireGLServer *server, WireGLClientConnection *client_connection )
{
	WireGLConnection *conn = (WireGLConnection *) wireGLAlloc( sizeof( *conn ) );
	socklen_t addr_length = sizeof( server->addr );
	listen( server->sock, 1 );

	conn->tcp_socket = accept( server->sock, (struct sockaddr *) &server->addr, &addr_length );
	if (conn->tcp_socket == -1)
	{
		wireGLSimpleError( "Accept error" );
	}

	conn->type = TCPIP;
	conn->wireGLReadMessage = wireGLTCPIPReadMessage;
	conn->wireGLSendMessage = wireGLTCPIPSendMessage;

	conn->wireGLReadMessage( conn, client_connection );
	conn->please_share_memory = client_connection->please_share_memory;
	conn->please_use_gm = client_connection->please_use_gm;
	return conn;
}	

UTIL_DECL void wireGLFinishAcceptClient( WireGLConnection *conn )
{
	conn->wireGLRegisterMemoryForDMA = wireGLNopRegisterMemoryForDMA;
	conn->wireGLDeregisterMemoryForDMA = wireGLNopDeregisterMemoryForDMA;

#ifdef GM_SUPPORT
	if (conn->please_use_gm)
		wireGLStartGM( conn );
#endif

	conn->receive_buf_one = (char *) wireGLAlloc( WIREGL_MAX_PACKET_SIZE );
	conn->receive_buf_two = (char *) wireGLAlloc( WIREGL_MAX_PACKET_SIZE );

	conn->wireGLRegisterMemoryForDMA( conn, conn->receive_buf_one, WIREGL_MAX_PACKET_SIZE );
	conn->wireGLRegisterMemoryForDMA( conn, conn->receive_buf_two, WIREGL_MAX_PACKET_SIZE );
}
