/* wgl.c
 *
 *  support for running applications under wireGL
 *
 *  Matthew Eldridge
 *  12-Oct-99
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifndef win32
#include <unistd.h>
#else
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <process.h>
#endif

static const char *progname = "wgl";
static const char *libgl    = NULL;
static int         verbose  = 0;

static const char *
basename( const char *path )
{
    char *last;
    last = strrchr( path, '/' );
    if ( !last )
        return path;
    else
        return last + 1;
}

static void *
xmalloc( size_t nbytes )
{
    void *ptr = malloc( nbytes );
    if ( ptr == NULL ) {
        fprintf( stderr, "%s: malloc( %d bytes ) failed!\n",
                 progname, nbytes );
        exit( 1 );
    }

    return ptr;
}

static void
setenv( const char *var, const char *value )
{
    unsigned long len;
    char *buf;

    len = strlen(var) + 1 + strlen(value) + 1;
    buf = (char *) malloc( len );
    sprintf( buf, "%s=%s", var, value );
    putenv( buf );

    if ( verbose )
        printf( "%s: %s\n", progname, buf );

    /* don't free the buf, the string is *part* of the environment,
     * and can't be reclaimed */
}

#ifdef win32

typedef struct TMPfile {
    char           *name;
    struct TMPfile *next;
} TMPfile;

typedef struct {
    char    *dirname;
    TMPfile *files;
} TMPdir;

static char *
error_string( int err )
{
    char *buf;

    FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER |
                   FORMAT_MESSAGE_FROM_SYSTEM, NULL, err,
                   MAKELANGID( LANG_NEUTRAL, SUBLANG_DEFAULT ),
                   (LPTSTR) &buf, 0, NULL );
    return buf;
}

char *
FindFileOnPath( const char *name )
{
    int   len;
    char *path, *tail;

    len = SearchPath( NULL, name, ".exe", 0, NULL, &tail );
    if ( len == 0 ) {
        int err = GetLastError( );
        fprintf( stderr, "%s: \"%s\" : %s\n",
                 progname, name, error_string( err ) );
        return NULL;
    }

	len+=2;
    path = (char *) xmalloc( len + 12 );

    if ( !SearchPath( NULL, name, ".exe", len, path, &tail ) ) {
        int err = GetLastError( );
        fprintf( stderr, "%s: \"%s\" : %s\n",
                 progname, name, error_string( err ) );
        return NULL;
    }

    return path;
}

int
CreateTempDir( TMPdir *tmpdir, char *pattern )
{
    char *name;

    tmpdir->dirname = NULL;
    tmpdir->files   = NULL;

    name = _tempnam( "C:\\", pattern );
    if ( name == NULL ) {
        fprintf( stderr, "%s: cannot create a unique directory name\n",
                 progname );
        return 0;
    }

    if ( verbose )
        printf( "%s: temp directory = \"%s\"\n", progname, name );

    if ( !CreateDirectory( name, NULL ) ) {
        int err = GetLastError( );
        fprintf( stderr, "%s: CreateDirectory( \"%s\" ) : %s\n",
                 progname, name, error_string( err ) );
        return 0;
    }

    tmpdir->dirname = name;
    return 1;
}

char *
CopyFileToTempDir( TMPdir *tmpdir, const char *filename )
{
    int len;
    char *fullpath, *tail;
    TMPfile *file;

    len = GetFullPathName( filename, 0, NULL, &tail );
    if ( len == 0 ) {
        int err = GetLastError( );
        fprintf( stderr, "%s: GetFullPathName( \"%s\" ): %s\n",
                 progname, filename, error_string( err ) );
        return NULL;
    }

    fullpath = (char *) xmalloc( len + 1 );
    if ( !GetFullPathName( filename, len, fullpath, &tail ) ) {
        int err = GetLastError( );
        fprintf( stderr, "%s: GetFullPathName( \"%s\" ): %s\n",
                 progname, filename, error_string( err ) );
        return NULL;
    }

    len = tail - fullpath;

    file = (TMPfile *) xmalloc( sizeof(*file) );
    file->name = (char *) xmalloc( strlen( tmpdir->dirname ) + 1 +
                                   strlen( tail ) + 1 );

    strcpy( file->name, tmpdir->dirname );
    strcat( file->name, "\\" );
    strcat( file->name, tail );

    if ( verbose )
        printf( "%s: copying \"%s\" -> \"%s\"\n", progname,
                fullpath, file->name );

    if ( !CopyFile( fullpath, file->name, 1 /* fail if exists */ ) ) {
        int err = GetLastError( );
        fprintf( stderr, "%s: copy \"%s\" -> \"%s\": %s\n",
                 progname, fullpath, file->name, error_string( err ) );
        return NULL;
    }

    file->next = tmpdir->files;
    tmpdir->files = file;

    return file->name;
}

void
DeleteTempDir( TMPdir *tmpdir )
{
    TMPfile *file;

    for ( file = tmpdir->files; file; file = file->next ) {

        if ( !DeleteFile( file->name ) ) {
            int err = GetLastError( );
            fprintf( stderr, "%s: DeleteFile( \"%s\" ): %s\n",
                     progname, file->name, error_string( err ) );
        }
    }

    if ( !RemoveDirectory( tmpdir->dirname ) ) {
        int err = GetLastError( );
        fprintf( stderr, "%s: RemoveDirectory( \"%s\" ): %s\n",
                 progname, tmpdir->dirname, error_string( err ) );
        fprintf( stderr, "%s: you must delete this directory by hand\n",
                 progname );
    }
}

static void
do_it( char *argv[] )
{
    TMPdir tmp;
    int status;

    if ( !CreateTempDir( &tmp, "wiregl" ) )
        exit( 1 );

    argv[0] = FindFileOnPath( argv[0] );
    if ( argv[0] == NULL )
        exit( 1 );

    argv[0] = CopyFileToTempDir( &tmp, argv[0] );
    if ( argv[0] == NULL )
        exit( 1 );

    if ( !CopyFileToTempDir( &tmp, libgl ) )
        exit( 1 );

    status = spawnv( _P_WAIT, argv[0], argv );

    if ( status == -1 ) {
        fprintf( stderr, "%s: \"%s\": %s\n", progname, argv[0],
                 strerror(errno) );
    } else {
        if ( status > 0 )
            fprintf( stderr, "%s: \"%s\": exited with status=%d\n",
                     progname, argv[0], status );
    }

    DeleteTempDir( &tmp );

    exit( status ? 1 : 0 );
}

#endif

#ifdef unix

static void
prefix_env_var( const char *prefix, const char *varname )
{
    char *val;

    val = getenv( varname );
    if ( val ) {
        unsigned long len = strlen( prefix ) + 1 + strlen( val ) + 1;
        char *buf = (char *) xmalloc( len );
        sprintf( buf, "%s:%s", prefix, val );
        setenv( varname, buf );
    } else {
        setenv( varname, prefix );
    }
}

static void
do_it( char *argv[] )
{
    if ( verbose )
        printf( "%s: libgl  = \"%s\"\n", progname, libgl );

    prefix_env_var( libgl, "LD_LIBRARY_PATH" );

#if defined(IRIX) || defined(IRIX64)
	/* if these are set then they'll be used by rld, so we have to
       make sure to modify them. */
	if ( getenv( "LD_LIBRARYN32_PATH" ) )
		prefix_env_var( libgl, "LD_LIBRARYN32_PATH" );

	if ( getenv( "LD_LIBRARY64_PATH" ) )
		prefix_env_var( libgl, "LD_LIBRARY64_PATH" );
#endif

    execvp( argv[0], argv );

    fprintf( stderr, "%s: \"%s\": %s\n", progname, argv[0], strerror(errno) );
    exit(1);
}

#endif

void
usage( void )
{
    char *libgl;

    fprintf( stderr,
             "usage: %s [options] <cmd> [args]\n"
             "options:\n"
#if win32
             "   -lib <file> : name of the replacement opengl32.dll\n"
#else
             "   -lib <dir>  : root of a directory tree containing\n"
             "                 {o32|n32|n64}/lib/libGL.so\n"
#endif
             "   -v[erbose]\n"
             "   -q[uiet]\n"
             "   -h[elp]     : this message\n"
             "\n"
             "The environment variable WIREGL_LIB specifies the default -lib,\n",
             progname );

    libgl = getenv( "WIREGL_LIB" );
    if ( libgl )
        fprintf( stderr, "currently set to \"%s\".\n", libgl );
    else
        fprintf( stderr, "currently not set.\n" );
}

int
main( int argc, char **argv )
{
    int i;

    progname = basename( argv[0] );

    libgl = NULL;

    for ( i = 1; i < argc; i++ ) {

        if ( argv[i][0] != '-' )
            break;

        if ( !strcmp( argv[i], "-lib" ) ) {

            i++;
            if ( i == argc ) {
                fprintf( stderr, "%s: %s expects argument\n",
                         progname, argv[i] );
                exit( 1 );
            }

            libgl = argv[i];
        }
        else if ( !strcmp( argv[i], "-v" ) ||
                  !strcmp( argv[i], "-verbose" ) ) {
            verbose = 1;
        }
        else if ( !strcmp( argv[i], "-q" ) ||
                  !strcmp( argv[i], "-quiet" ) ) {
            verbose = 0;
        }
        else if ( !strcmp( argv[i], "-h" ) ||
                  !strcmp( argv[i], "-help" ) ) {

            usage( );
            exit( 0 );
        }
        else {

            fprintf( stderr, "%s: unknown argument: %s\n", progname, argv[i] );
            usage( );
            exit( 1 );
        }
    }

    argc -= i;
    argv += i;

    if ( argc < 1 ) {
        fprintf( stderr, "%s: no command to run?\n", progname );
        exit( 1 );
    }

    if ( libgl == NULL ) {
        libgl = getenv( "WIREGL_LIB" );
        if ( libgl == NULL ) {
            fprintf( stderr, "%s: must specify -lib or WIREGL_LIB\n",
                     progname );
            exit( 1 );
        }
    }

    do_it( argv );

    return 0;
}
